from __future__ import annotations

import ibis
import pytest

from mismo.arrays import array_any
from mismo.compare import LevelComparer, MatchLevel


class EmailMatchLevel(MatchLevel):
    FULL_EXACT = 0
    FULL_NEAR = 1
    LOCAL_EXACT = 2
    LOCAL_NEAR = 3
    ELSE = 4
    _UNDERED = 4
    NOT_INT = "foo"

    def ignored(self):
        pass


def test_constructors():
    assert EmailMatchLevel(1).as_integer() is 1  # noqa: F632
    assert EmailMatchLevel("FULL_NEAR").as_integer() is 1  # noqa: F632
    with pytest.raises(TypeError):
        EmailMatchLevel(None)
    with pytest.raises(ValueError):
        EmailMatchLevel(100)
    with pytest.raises(ValueError):
        EmailMatchLevel("full_near")


def test_getitem():
    assert isinstance(EmailMatchLevel[1], str)
    assert EmailMatchLevel[1] == "FULL_NEAR"

    assert isinstance(EmailMatchLevel["FULL_NEAR"], int)
    assert EmailMatchLevel["FULL_NEAR"] == 1

    with pytest.raises(TypeError):
        EmailMatchLevel[None]
    with pytest.raises(KeyError):
        EmailMatchLevel[100]
    with pytest.raises(KeyError):
        EmailMatchLevel["full_near"]

    assert EmailMatchLevel[ibis.literal(1)].execute() == "FULL_NEAR"
    assert EmailMatchLevel[ibis.literal("FULL_NEAR")].execute() == 1

    assert EmailMatchLevel[ibis.literal(100)].execute() is None
    assert EmailMatchLevel[ibis.literal("full_near")].execute() is None


def test_container_semantics():
    assert len(EmailMatchLevel) == 5
    assert "FULL_EXACT" in EmailMatchLevel
    assert "full_exact" not in EmailMatchLevel
    assert 0 in EmailMatchLevel
    assert -1 not in EmailMatchLevel
    with pytest.raises(TypeError):
        assert bool not in EmailMatchLevel
    with pytest.raises(TypeError):
        ibis.literal(0) in EmailMatchLevel
    with pytest.raises(TypeError):
        ibis.literal("FULL_EXACT") in EmailMatchLevel


def test_eq_neq():
    assert EmailMatchLevel.FULL_EXACT == 0
    assert EmailMatchLevel.FULL_EXACT != 1
    assert EmailMatchLevel("FULL_EXACT") == 0
    assert EmailMatchLevel("FULL_EXACT") != 1
    assert EmailMatchLevel(0) == 0
    assert EmailMatchLevel(0) != 1

    assert EmailMatchLevel.FULL_EXACT == "FULL_EXACT"
    assert EmailMatchLevel.FULL_EXACT != "full_exact"
    assert EmailMatchLevel("FULL_EXACT") == "FULL_EXACT"
    assert EmailMatchLevel("FULL_EXACT") != "full_exact"
    assert EmailMatchLevel(0) == "FULL_EXACT"
    assert EmailMatchLevel(0) != "full_exact"

    assert EmailMatchLevel.FULL_EXACT == EmailMatchLevel.FULL_EXACT
    assert EmailMatchLevel.FULL_EXACT != EmailMatchLevel.FULL_NEAR
    assert EmailMatchLevel("FULL_EXACT") == EmailMatchLevel.FULL_EXACT
    assert EmailMatchLevel("FULL_EXACT") != EmailMatchLevel.FULL_NEAR
    assert EmailMatchLevel(0) == EmailMatchLevel.FULL_EXACT
    assert EmailMatchLevel(0) != EmailMatchLevel.FULL_NEAR


def test_ordering_not_supported():
    with pytest.raises(TypeError):
        EmailMatchLevel.FULL_EXACT < 1
    with pytest.raises(TypeError):
        EmailMatchLevel.FULL_EXACT <= 1
    with pytest.raises(TypeError):
        EmailMatchLevel.FULL_EXACT > 1
    with pytest.raises(TypeError):
        EmailMatchLevel.FULL_EXACT >= 1

    with pytest.raises(TypeError):
        EmailMatchLevel("FULL_EXACT") < 1
    with pytest.raises(TypeError):
        EmailMatchLevel("FULL_EXACT") <= 1
    with pytest.raises(TypeError):
        EmailMatchLevel("FULL_EXACT") > 1
    with pytest.raises(TypeError):
        EmailMatchLevel("FULL_EXACT") >= 1


def test_repr():
    assert (
        repr(EmailMatchLevel)
        == "EmailMatchLevel(FULL_EXACT=0, FULL_NEAR=1, LOCAL_EXACT=2, LOCAL_NEAR=3, ELSE=4)"  # noqa: E501
    )
    assert repr(EmailMatchLevel.FULL_EXACT) == "EmailMatchLevel.FULL_EXACT"
    assert repr(EmailMatchLevel("FULL_EXACT")) == "EmailMatchLevel.FULL_EXACT"
    assert repr(EmailMatchLevel(0)) == "EmailMatchLevel.FULL_EXACT"

    ibis_repr = "EmailMatchLevel(<foo>)"
    assert repr(EmailMatchLevel(ibis.literal("FULL_EXACT").name("foo"))) == ibis_repr
    assert repr(EmailMatchLevel(ibis.literal(0).name("foo"))) == ibis_repr


def test_str():
    # class
    assert (
        str(EmailMatchLevel)
        == "EmailMatchLevel(FULL_EXACT=0, FULL_NEAR=1, LOCAL_EXACT=2, LOCAL_NEAR=3, ELSE=4)"  # noqa: E501
    )
    # instance
    assert str(EmailMatchLevel.FULL_EXACT) == "FULL_EXACT"
    assert str(EmailMatchLevel("FULL_EXACT")) == "FULL_EXACT"
    assert str(EmailMatchLevel(0)) == "FULL_EXACT"
    with pytest.raises(TypeError):
        str(EmailMatchLevel(ibis.literal("FULL_EXACT")))
    with pytest.raises(TypeError):
        str(EmailMatchLevel(ibis.literal(0)))


def test_int():
    # class
    with pytest.raises(TypeError):
        int(EmailMatchLevel)
    # instance
    assert int(EmailMatchLevel.FULL_EXACT) == 0
    assert int(EmailMatchLevel("FULL_EXACT")) == 0
    assert int(EmailMatchLevel(0)) == 0
    with pytest.raises(TypeError):
        int(EmailMatchLevel(ibis.literal("FULL_EXACT")))
    with pytest.raises(TypeError):
        int(EmailMatchLevel(ibis.literal(0)))


def test_conversion():
    assert EmailMatchLevel.FULL_NEAR.as_integer() == 1
    assert EmailMatchLevel.FULL_NEAR.as_string() == "FULL_NEAR"

    assert EmailMatchLevel(1).as_integer() == 1
    assert EmailMatchLevel(1).as_string() == "FULL_NEAR"
    assert EmailMatchLevel("FULL_NEAR").as_integer() == 1
    assert EmailMatchLevel("FULL_NEAR").as_string() == "FULL_NEAR"

    ei = ibis.literal(1)
    es = ibis.literal("FULL_NEAR")
    assert EmailMatchLevel(ei).as_integer().execute() == 1
    assert EmailMatchLevel(ei).as_string().execute() == "FULL_NEAR"
    assert EmailMatchLevel(es).as_integer().execute() == 1
    assert EmailMatchLevel(es).as_string().execute() == "FULL_NEAR"


def test_match_levels_bad():
    with pytest.raises(TypeError):

        class DuplicateValues(MatchLevel):
            X = 1
            Y = 1


def test_array_based_conditions(table_factory):
    class TagMatchLevel(MatchLevel):
        ANY_EQUAL = 0
        ELSE = 1

    tag_comparer = LevelComparer(
        name="tag",
        levels=TagMatchLevel,
        cases=[
            (
                lambda pairs: array_any(
                    pairs.tags_l.map(
                        lambda ltag: pairs.tags_r.map(lambda rtag: ltag == rtag)
                    ).flatten(),
                ),
                TagMatchLevel.ANY_EQUAL,
            ),
            (True, TagMatchLevel.ELSE),
        ],
        representation="string",
    )

    t = table_factory(
        {
            "tags_l": [["a", "b"], ["c", "d"], [], None],
            "tags_r": [["b", "x"], ["y", "z"], ["m"], []],
            "expected": ["ANY_EQUAL", "ELSE", "ELSE", "ELSE"],
        }
    )
    t = tag_comparer(t)
    assert t.tag.execute().tolist() == t.expected.execute().tolist()



from __future__ import annotations

import pytest

from mismo.lib.name import NameBlocker


def test_name_blocker_with_column(name_table):
    blocker = NameBlocker(column="name")
    linkage = blocker(name_table, name_table)
    record_ids = linkage.links.select("record_id_l", "record_id_r").execute()
    record_ids = set(
        [tuple(pair) for pair in record_ids.itertuples(index=False, name=None)]
    )
    expected = {
        ("alice_anderson", "a_anderson_phd"),
        ("bob_baker", "robert_b_baker_jr"),
        ("charles_carter", "mr_charles_carter"),
    }
    expected = set((a, b) if a < b else (b, a) for a, b in expected)
    assert record_ids == expected


def test_name_blocker_with_invalid_arguments():
    with pytest.raises(ValueError):
        NameBlocker()
    with pytest.raises(ValueError):
        NameBlocker(column="name", column_left="name", column_right="name")
        NameBlocker(column="name", column_left="name")
    with pytest.raises(ValueError):
        NameBlocker(column_left="name")
    with pytest.raises(ValueError):
        NameBlocker(column_right="name")



from __future__ import annotations

import ibis
import pytest

from mismo import text


@pytest.mark.parametrize(
    "inp,n,exp",
    [
        ("abc", 2, ["ab", "bc"]),
        ("abcd", 2, ["ab", "bc", "cd"]),
        ("abc", 3, ["abc"]),
        ("abcd", 3, ["abc", "bcd"]),
        (
            "abcdef",
            3,
            ["abc", "bcd", "cde", "def"],
        ),
        ("", 2, []),
        ("a", 2, []),
        (None, 4, None),
    ],
)
def test_ngrams(inp, n, exp):
    result = text.ngrams(inp, n).execute()
    if exp is None:
        assert result is None
    else:
        assert set(result) == set(exp)


@pytest.mark.parametrize(
    "inp,exp",
    [
        ("abc", ["abc"]),
        ("abc def", ["abc", "def"]),
        ("abc  def", ["abc", "def"]),
        ("  abc  def", ["abc", "def"]),
        (" ", []),
        ("", []),
        (None, None),
    ],
)
def test_tokenize(inp, exp):
    result = text.tokenize(ibis.literal(inp, type=str)).execute()
    if exp is None:
        assert result is None
    else:
        assert result == exp


@pytest.mark.parametrize(
    "inp,expected",
    [
        ("müller", "muller"),
        ("Muñoz", "Munoz"),
        ("François", "Francois"),
        ("pérez", "perez"),
        ("à", "a"),
        ("á", "a"),
        ("â", "a"),
        ("ä", "a"),
        ("ã", "a"),
        ("å", "a"),
        ("Žižek", "Zizek"),
        ("Øslo", "Øslo"),  # unchanged
        ("østergaard", "østergaard"),  # unchanged
        ("æ", "æ"),  # unchanged
        ("ɑɽⱤoW", "ɑɽⱤoW"),  # unchanged
        ("aAaz;ZæÆ&", "aAaz;ZæÆ&"),  # unchanged
        ("ıI", "ıI"),  # unchanged
        ("", ""),
        (None, None),
    ],
)
def test_strip_accents(inp, expected):
    result = text.strip_accents(ibis.literal(inp, type="string")).execute()
    assert expected == result



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ibis.expr import types as ir

from mismo._util import optional_import

if TYPE_CHECKING:
    import numpy as np


def adjusted_mutual_info_score(
    labels_true: ir.Table,
    labels_pred: ir.Table,
    *,
    average_method: str = "arithmetic",
) -> float:
    """Adjusted Mutual Information between two clusterings.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.adjusted_mutual_info_score][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.adjusted_mutual_info_score(
        labels_true, labels_pred, average_method=average_method
    )


def adjusted_rand_score(labels_true: ir.Table, labels_pred: ir.Table) -> float:
    """Adjusted Rand Index between two clusterings.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.adjusted_rand_score][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.adjusted_rand_score(labels_true, labels_pred)


def fowlkes_mallows_score(labels_true: ir.Table, labels_pred: ir.Table) -> float:
    """Measure the similarity of two clusterings of a set of points.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.fowlkes_mallows_score][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.fowlkes_mallows_score(labels_true, labels_pred)


def completeness_score(labels_true: ir.Table, labels_pred: ir.Table) -> float:
    """Compute completeness metric of a cluster labeling given a ground truth.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.completeness_score][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.completeness_score(labels_true, labels_pred)


def homogeneity_score(labels_true: ir.Table, labels_pred: ir.Table) -> float:
    """Homogeneity metric of a cluster labeling given a ground truth.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.homogeneity_score][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.homogeneity_score(labels_true, labels_pred)


def v_measure_score(labels_true: ir.Table, labels_pred: ir.Table) -> float:
    """V-measure metric of a cluster labeling given a ground truth.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.v_measure_score][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.v_measure_score(labels_true, labels_pred)


def homogeneity_completeness_v_measure(
    labels_true: ir.Table, labels_pred: ir.Table, *, beta: float = 1.0
) -> tuple[float, float, float]:
    """Compute the homogeneity, completeness, and V-measure scores at once.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.homogeneity_completeness_v_measure][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.homogeneity_completeness_v_measure(
        labels_true, labels_pred, beta=beta
    )


def mutual_info_score(
    labels_true: ir.Table, labels_pred: ir.Table, *, contingency: Any = None
) -> float:
    """Compute the mutual information between two clusterings.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.mutual_info_score][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.mutual_info_score(labels_true, labels_pred, contingency=contingency)


def normalized_mutual_info_score(
    labels_true: ir.Table, labels_pred: ir.Table, *, average_method: str = "arithmetic"
) -> float:
    """Compute the normalized mutual information between two clusterings.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.normalized_mutual_info_score][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.normalized_mutual_info_score(
        labels_true, labels_pred, average_method=average_method
    )


def rand_score(labels_true: ir.Table, labels_pred: ir.Table) -> float:
    """Compute the Rand Index between two clusterings.

    The two input tables must have columns "record_id" and "label",
    a map from record ID to cluster label. They must have the same record IDs.

    See [sklearn.metrics.rand_score][] for more information.
    """
    with optional_import("scikit-learn"):
        from sklearn import metrics as _metrics
    labels_true, labels_pred = _to_numpy_labels(labels_true, labels_pred)
    return _metrics.rand_score(labels_true, labels_pred)


def _to_numpy_labels(
    labels_true: ir.Table, labels_pred: ir.Table
) -> tuple[np.ndarray, np.ndarray]:
    labels_true = labels_true["record_id", "label"]
    labels_pred = labels_pred["record_id", "label"]
    labels_true_df = labels_true.order_by("record_id").to_pandas()
    labels_pred_df = labels_pred.order_by("record_id").to_pandas()
    if not labels_true_df.record_id.equals(labels_pred_df.record_id):
        raise ValueError("labels_true and labels_pred must be aligned")
    return labels_true_df.label.values, labels_pred_df.label.values



from __future__ import annotations

import inspect

from mismo import exceptions
from mismo.exceptions import MismoError, MismoWarning


def test_mismo_error_inheritance():
    """Test that MismoError is a base Exception."""
    assert issubclass(MismoError, Exception)


def test_mismo_warning_inheritance():
    """Test that MismoWarning is a base Warning."""
    assert issubclass(MismoWarning, Warning)


def test_all_errors_inherit_from_mismo_error():
    """Test that all exception classes inherit from MismoError."""
    error_classes = [
        cls
        for name, cls in inspect.getmembers(exceptions, inspect.isclass)
        if cls.__module__ == exceptions.__name__
        and issubclass(cls, Exception)
        and not issubclass(cls, Warning)  # Exclude warnings
        and cls is not MismoError
    ]
    assert len(error_classes) > 0, "No error classes found in mismo.exceptions"
    for error_class in error_classes:
        assert issubclass(error_class, MismoError)


def test_all_warnings_inherit_from_mismo_warning():
    """Test that all warning classes inherit from MismoWarning."""
    warning_classes = [
        cls
        for name, cls in inspect.getmembers(exceptions, inspect.isclass)
        if cls.__module__ == exceptions.__name__
        and issubclass(cls, Warning)
        and cls is not MismoWarning
    ]
    assert len(warning_classes) > 0, "No warning classes found in mismo.exceptions"
    for warning_class in warning_classes:
        assert issubclass(warning_class, MismoWarning)



from __future__ import annotations

import warnings

import ibis

from mismo.types import LinksTable


# TODO: could we come up with an algorithm that would allow for a random seed?
def sample_all_links(
    left: ibis.Table,
    right: ibis.Table,
    *,
    max_pairs: int | None = None,
) -> LinksTable:
    """Samples up to `max_pairs` from all possible pairs of records.

    Parameters
    ----------
    left :
        The left table.
    right :
        The right table.
    max_pairs :
        The maximum number of pairs to sample. If None, all possible pairs are sampled.

    Returns
    -------
    A [LinksTable][mismo.LinksTable] with just record_id_l and record_id_r.
    All pairs will be unique.

    Examples
    --------
    >>> import ibis
    >>> import mismo
    >>> ibis.options.interactive = True
    >>> t, _labels = mismo.playdata.load_febrl1()
    >>> t.head(5)
    ┏━━━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━┳━━━┓
    ┃ record_id    ┃ given_name ┃ surname    ┃ street_number ┃ address_1         ┃ … ┃
    ┡━━━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━╇━━━┩
    │ string       │ string     │ string     │ string        │ string            │ … │
    ├──────────────┼────────────┼────────────┼───────────────┼───────────────────┼───┤
    │ rec-0-dup-0  │ thomas     │ rokobaro   │ 12            │ herschell circuit │ … │
    │ rec-0-org    │ flynn      │ rokobaro   │ 12            │ herschell circuit │ … │
    │ rec-1-dup-0  │ karli      │ alderson   │ 144           │ nulsen circuit    │ … │
    │ rec-1-org    │ karli      │ alderson   │ 144           │ nulsen circuit    │ … │
    │ rec-10-dup-0 │ kayla      │ harrington │ NULL          │ maltby circuit    │ … │
    └──────────────┴────────────┴────────────┴───────────────┴───────────────────┴───┘
    >>> mismo.linkage.sample_all_links(t, t, max_pairs=7)  # doctest: +SKIP
    ┏━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━┓
    ┃ record_id_l   ┃ record_id_r   ┃
    ┡━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━┩
    │ string        │ string        │
    ├───────────────┼───────────────┤
    │ rec-226-dup-0 │ rec-396-org   │
    │ rec-232-dup-0 │ rec-402-dup-0 │
    │ rec-259-dup-0 │ rec-61-org    │
    │ rec-293-dup-0 │ rec-41-dup-0  │
    │ rec-448-org   │ rec-25-org    │
    └───────────────┴───────────────┘
    """  # noqa: E501
    left = left.cache()
    right = right.cache()
    n_possible_pairs = int(left.count().execute() * right.count().execute())
    n_pairs = (
        n_possible_pairs if max_pairs is None else min(n_possible_pairs, max_pairs)
    )
    if n_pairs > 100_000_000:
        msg = f"{n_pairs:,}" if n_pairs != n_possible_pairs else "all"
        warnings.warn(
            f"Sampling {msg} pairs from {n_possible_pairs:,} possible pairs."
            " This may be slow. Consider setting max_pairs to a smaller value."
        )
    if max_pairs is None:
        return LinksTable.from_join_condition(left, right, True)

    def make_pair_ids():
        pair_ids = ibis.range(n_pairs).unnest().as_table()
        return pair_ids.select(
            __left_id=(ibis.random() * left.count()).floor().cast("int64"),
            __right_id=(ibis.random() * right.count()).floor().cast("int64"),
        )

    # This iterative method is the best I could come up with to avoid duplicates.
    # A better method might be possible.
    # At first I tried ibis.range(n_pairs * 2)....distinct().limit(n_pairs)
    # but choosing the safety margin (here it is 2) is tricky.
    # The closer n_pairs gets to n_possible_pairs, the higher the safety margin
    # needs to be.
    # But you are doing unneeded computation if you set it higher than it needs
    # to be.
    # For example if we have 10 records, and max_pairs is 99, then we need
    # to do really large samples in order to get to the 99 unique pairs.
    # On the other hand, if we have 10k records and max_pairs is only 10,
    # then we could get away with doing a sample of 20 pairs and we would
    # probably get all unique pairs.
    pair_ids = make_pair_ids().distinct().cache()
    while pair_ids.count().execute() < n_pairs:
        pair_ids = pair_ids.union(make_pair_ids()).distinct().limit(n_pairs).cache()

    right = right.view()
    left_ids = left.select(record_id_l="record_id").mutate(__left_id=ibis.row_number())
    right_ids = right.select(record_id_r="record_id").mutate(
        __right_id=ibis.row_number()
    )
    raw_links = (
        pair_ids.left_join(left_ids, "__left_id")
        .left_join(right_ids, "__right_id")
        .drop("__left_id", "__right_id", "__left_id_right", "__right_id_right")
    )
    return LinksTable(raw_links, left=left, right=right)



from __future__ import annotations

import ibis
from ibis.common.exceptions import IbisTypeError
import pytest

import mismo
from mismo import joins
from mismo.joins import _conditions


@pytest.mark.parametrize(
    "obj",
    [
        pytest.param("foo", id="str"),
        pytest.param(ibis._.foo, id="deferred"),
    ],
)
def test_key_join_condition(obj):
    condition = joins.join_condition(obj)
    assert isinstance(condition, _conditions.KeyJoinCondition)
    t1 = ibis.table({"foo": "int64"}, name="t1")
    t2 = ibis.table({"foo": "int64"}, name="t2")
    assert condition.__join_condition__(t1, t2).equals(t1.foo == t2.foo)

    tbad = ibis.table({"bar": "int64"}, name="tbad")
    with pytest.raises((IbisTypeError, AttributeError)):
        condition.__join_condition__(t1, tbad)


def test_deferred_left_right():
    condition = joins.join_condition(mismo.left.foo == mismo.right.bar)
    assert isinstance(condition, _conditions.LeftRightDeferredCondition)
    t1 = ibis.table({"foo": "int64"}, name="t1")
    t2 = ibis.table({"bar": "int64"}, name="t2")
    assert condition.__join_condition__(t1, t2).equals(t1.foo == t2.bar)

    tbad = ibis.table({"baz": "int64"}, name="tbad")
    with pytest.raises(AttributeError):
        condition.__join_condition__(t1, tbad)


def test_deferred_bogus():
    with pytest.raises(NotImplementedError):
        joins.join_condition(mismo.left.foo == ibis._.bar)
    with pytest.raises(NotImplementedError):
        joins.join_condition(mismo.left.foo == "hello")



from __future__ import annotations

import ibis


def check_tables_and_links(
    left: ibis.Table, right: ibis.Table, links: ibis.Table
) -> None:
    if "record_id" not in left.columns:
        raise ValueError("column 'record_id' not in table")
    if "record_id" not in right.columns:
        raise ValueError("column 'record_id' not in other")
    if "record_id_l" not in links.columns:
        raise ValueError("column 'record_id_l' not in links")
    if "record_id_r" not in links.columns:
        raise ValueError("column 'record_id_r' not in links")
    try:
        left.record_id == links.record_id_l
    except Exception:
        raise ValueError(
            f"left.record_id of type {left.record_id.type()} is not comparable with links.record_id_l of type {links.record_id_l.type()}"  # noqa: E501
        )
    try:
        right.record_id == links.record_id_r
    except Exception:
        raise ValueError(
            f"right.record_id of type {right.record_id.type()} is not comparable with links.record_id_r of type {links.record_id_r.type()}"  # noqa: E501
        )



from __future__ import annotations

from abc import ABCMeta
from typing import Iterable, Iterator, Literal, Type, TypeVar, overload

import ibis
from ibis.expr import types as ir

from mismo import _typing, _util


class _LevelsMeta(ABCMeta):
    def __new__(cls, name, bases, dct):
        s2i = {
            k: v for k, v in dct.items() if not k.startswith("_") and isinstance(v, int)
        }
        val_to_names = {v: set() for v in s2i.values()}
        for k, v in s2i.items():
            val_to_names[v].add(k)
        dupes = {}
        for v, names in val_to_names.items():
            if len(names) > 1:
                for n in names:
                    dupes[n] = v
        if dupes:
            raise TypeError(f"Duplicate values: {dupes}")
        i2s = {i: s for s, i in s2i.items()}

        newcls = super().__new__(cls, name, bases, dct)
        newcls.__s2i__ = s2i
        newcls.__i2s__ = i2s
        newcls.__annotations__ = {
            **_typing.get_annotations(newcls),
            **{k: newcls for k in s2i},
        }
        for s in s2i:
            setattr(newcls, s, newcls(s))
        return newcls

    @overload
    def __getitem__(self, key: str) -> int: ...

    @overload
    def __getitem__(self, key: int) -> str: ...

    @overload
    def __getitem__(self, key: ir.NumericValue) -> ir.StringValue: ...

    @overload
    @classmethod
    def __getitem__(self, key: ir.StringValue) -> ir.IntegerValue: ...

    def __getitem__(self, key):
        """Converts string-likes to int-likes, and vice-versa.

        Ibis expressions are converted, but if you supply a bad value like
        "bogus level name", we can't type-check you at call time,
        and down the line when you actually .execute() the results you will get an
        ibis NULL value.

        Examples
        --------
        >>> from mismo.compare import MatchLevel
        >>> class NameMatchLevel(MatchLevel):
        ...     EXACT = 0
        ...     NEAR = 1
        ...     ELSE = 2

        >>> NameMatchLevel[1]
        'NEAR'
        >>> NameMatchLevel["NEAR"]
        1
        >>> NameMatchLevel[ibis.literal(1)].execute()
        'NEAR'
        >>> NameMatchLevel[ibis.literal("NEAR")].execute()
        1
        >>> NameMatchLevel[100]
        Traceback (most recent call last):
            ...
        KeyError: Invalid value: 100. Must be one of {0, 1, 2}
        >>> NameMatchLevel[ibis.literal(100)].execute() is None
        True
        """
        if _is_stringy(key):
            return self._str_to_int(key)
        elif _is_inty(key):
            return self._int_to_str(key)
        else:
            raise TypeError(f"Invalid key: {key}")

    def __contains__(self, key: str | int) -> bool:
        """Check if an int or str is one of the levels."""
        if not isinstance(key, (str, int)):
            raise TypeError(f"can only check containment for ints and strs, got {key}")
        return key in self.__s2i__ or key in self.__i2s__

    def __iter__(self) -> Iterator[str]:
        """The names of the levels."""
        return iter(self.__s2i__.keys())

    def __len__(self) -> int:
        """The number of levels"""
        return len(self.__s2i__)

    def __repr__(self) -> str:
        pairs = (f"{k}={v}" for k, v in self.__s2i__.items())
        return f"{self.__name__}({', '.join(pairs)})"


class MatchLevel(metaclass=_LevelsMeta):
    """An enum-like class for match levels.

    This class is used to define the levels of agreement between two records.

    Examples
    --------
    >>> from mismo.compare import MatchLevel
    >>> class NameMatchLevel(MatchLevel):
    ...     EXACT = 0
    ...     NEAR = 1
    ...     ELSE = 2

    The class acts as a container:

    >>> len(NameMatchLevel)
    3
    >>> 2 in NameMatchLevel
    True
    >>> list(NameMatchLevel)
    ['EXACT', 'NEAR', 'ELSE']

    You can access the hardcoded values:

    >>> str(NameMatchLevel.EXACT)
    'EXACT'
    >>> int(NameMatchLevel.EXACT)
    0

    You can use indexing semantics to translate between strings and ints:

    >>> NameMatchLevel[1]
    'NEAR'
    >>> NameMatchLevel["NEAR"]
    1
    >>> NameMatchLevel[ibis.literal(1)].execute()
    'NEAR'
    >>> NameMatchLevel[ibis.literal("NEAR")].execute()
    1

    You can construct your own values, both from python literals...

    >>> NameMatchLevel("NEAR").as_integer()
    1
    >>> NameMatchLevel(2).as_string()
    'ELSE'
    >>> NameMatchLevel(3)
    Traceback (most recent call last):
    ...
    ValueError: Invalid value: 3. Must be one of {0, 1, 2}`

    ...And Ibis expressions

    >>> import ibis
    >>> levels_raw = ibis.array([0, 2, 1, 99]).unnest()
    >>> levels = NameMatchLevel(levels_raw)
    >>> levels.as_string().execute()
    0    EXACT
    1     ELSE
    2     NEAR
    3     None
    Name: NameMatchLevel, dtype: object
    >>> levels.as_integer().name("levels").execute()
    0     0
    1     2
    2     1
    3    99
    Name: levels, dtype: int8

    Comparisons work as you expect:

    >>> NameMatchLevel.NEAR == 1
    True
    >>> NameMatchLevel(1) == "NEAR"
    True
    >>> (levels_raw == NameMatchLevel.NEAR).name("eq").execute()
    0    False
    1    False
    2     True
    3    False
    Name: eq, dtype: bool

    However, implicit ordering is not supported
    (file an issue if you think it should be):

    >>> NameMatchLevel.NEAR > 0
    Traceback (most recent call last):
    ...
    TypeError: '>' not supported between instances of 'NameMatchLevel' and 'int'
    """

    def __init__(
        self, value: MatchLevel | int | str | ir.StringValue | ir.IntegerValue
    ):
        """Create a new match level value.

        If the given value is a python int or str, it is checked against the
        valid values for this class. If it is an ibis expression,
        we do no such check.

        Parameters
        ----------
        value :
            The value of the match level.
        """
        if not isinstance(
            value, (MatchLevel, ir.StringValue, ir.IntegerValue, int, str)
        ):
            raise TypeError(f"Invalid value: {value}")
        if isinstance(value, str):
            self._check_str(value, ValueError)
        elif isinstance(value, int):
            self._check_int(value, ValueError)
        if isinstance(value, MatchLevel):
            self._value = value._value
        else:
            self._value = value

    def as_integer(self) -> int | ir.IntegerValue:
        """Convert to a python int or ibis integer, depending on the original type."""
        if _is_inty(self._value):
            return self._value
        return self._str_to_int(self._value)

    def as_string(self) -> str | ir.StringValue:
        """Convert to a python str or ibis string, depending on the original type."""
        if _is_stringy(self._value):
            return self._value
        return self._int_to_str(self._value)

    def __repr__(self) -> str:
        if _is_python(self._value):
            return self.__class__.__name__ + "." + str(self)
        else:
            return self.__class__.__name__ + "(<" + self._value.get_name() + ">)"

    def __str__(self) -> str:
        if not _is_python(self._value):
            raise TypeError(
                "You can only only use str() on MatchLevels made from str's and int's. "
                f"This {self.__class__.__name__} is made of {self._value.__class__.__name__}"  # noqa: E501
            )
        return self.as_string()

    def __int__(self) -> int:
        if not _is_python(self._value):
            raise TypeError(
                "You can only only use int() on MatchLevels made from str's and int's. "
                f"This {self.__class__.__name__} is made of {self._value.__class__.__name__}"  # noqa: E501
            )
        return self.as_integer()

    def __eq__(
        self, other: int | str | ir.NumericValue | ir.StringValue | MatchLevel
    ) -> bool | ir.BooleanValue:
        """"""
        if isinstance(other, MatchLevel):
            other = other._value
        if _is_inty(other):
            return self.as_integer() == other
        elif _is_stringy(other):
            return self.as_string() == other
        else:
            return NotImplemented

    @classmethod
    def _str_to_int(cls, v: str | ir.StringValue) -> int | ir.IntegerValue:
        # assumes v is in __s2i__
        if isinstance(v, str):
            cls._check_str(v, KeyError)
            return cls.__s2i__[v]
        elif isinstance(v, ir.StringValue):
            return v.substitute(cls.__s2i__, else_=ibis.null()).name(cls.__name__)
        assert False

    @classmethod
    def _int_to_str(cls, v: int | ir.IntegerValue) -> str | ir.StringValue:
        # assumes v is in __i2s__
        if isinstance(v, int):
            cls._check_int(v, KeyError)
            return cls.__i2s__[v]
        elif isinstance(v, ir.IntegerValue):
            return v.substitute(cls.__i2s__, else_=ibis.null()).name(cls.__name__)
        assert False

    @classmethod
    def _check_int(cls, value: int, err_type) -> None:
        valid_ints = set(cls.__i2s__.keys())
        if value not in valid_ints:
            raise err_type(f"Invalid value: {value}. Must be one of {valid_ints}")

    @classmethod
    def _check_str(cls, value: str, err_type) -> None:
        valid_strs = set(cls.__s2i__.keys())
        if value not in valid_strs:
            raise err_type(f"Invalid value: {value}. Must be one of {valid_strs}")


def _is_python(v):
    return isinstance(v, (int, str))


def _is_inty(v):
    return isinstance(v, (int, ir.NumericValue))


def _is_stringy(v):
    return isinstance(v, (str, ir.StringValue))


MatchLevelT = TypeVar("MatchLevelT", bound=MatchLevel)


class LevelComparer:
    """
    Assigns a MatchLevel to record pairs based on one dimension, e.g. *name*
    """

    def __init__(
        self,
        name: str,
        levels: Type[MatchLevelT],
        cases: Iterable[tuple[ir.BooleanColumn, MatchLevelT]],
        *,
        representation: Literal["string", "integer"] = "integer",
    ):
        self.name = name
        self.levels = levels
        self.cases = tuple((c, self.levels(lev)) for c, lev in cases)
        self.representation = representation

    name: str
    """The name of the comparer, eg "date", "address", "latlon", "price"."""
    levels: Type[MatchLevelT]
    """The levels of agreement."""
    cases: tuple[tuple[ir.BooleanColumn, MatchLevelT], ...]
    """The cases to check for each level."""
    representation: Literal["string", "integer"] = "integer"
    """The native representation of the levels in ibis expressions.

    Integers are more performant, but strings are more human-readable.
    """

    def __call__(
        self,
        pairs: ir.Table,
        *,
        representation: Literal["string", "integer"] | None = None,
    ) -> ir.StringColumn | ir.IntegerColumn:
        """Label each record pair with the level that it matches.

        Go through the levels in order. If a record pair matches a level, label ir.
        If none of the levels match a pair, it labeled as "else".

        Parameters
        ----------
        pairs : Table
            A table of record pairs.
        Returns
        -------
        labels : StringColumn
            The labels for each record pair.
        """
        if representation is None:
            representation = self.representation

        cases = [(pairs.bind(c)[0], level) for c, level in self.cases]
        if representation == "string":
            cases = [(c, level.as_string()) for c, level in cases]
        elif representation == "integer":
            cases = [(c, level.as_integer()) for c, level in cases]
        else:
            raise ValueError(f"Invalid representation: {representation}")

        return pairs.mutate(_util.cases(*cases).name(self.name))

    def __repr__(self) -> str:
        return f"LevelComparer(name={self.name}, levels=[{self.levels}])"



from __future__ import annotations

import pytest

from mismo.lib import geo


@pytest.mark.parametrize(
    "addresses, expected",
    [
        pytest.param(
            [
                {
                    "street1": "132 Main St",
                    "street2": "Apt    1-b",
                    "city": "Springfield",
                    "state": "",
                    "postal_code": "62701",
                    "country": "US",
                },
                {
                    "street1": "1 1st Ave",
                    "street2": None,
                    "city": "Springfield",
                    "state": "",
                    "postal_code": "62701",
                },
            ],
            {
                "addresses": [
                    {
                        "is_pobox": False,
                        "street1": "132 MAIN ST",
                        "city": "SPRINGFIELD",
                        "state": None,
                        "postal_code": "62701",
                        "street_trigrams": [
                            "132",
                            "32 ",
                            "2 M",
                            " MA",
                            "MAI",
                            "AIN",
                            "IN ",
                            "N S",
                            " ST",
                        ],
                        "street_name": "MAIN",
                        "street_number": "132",
                        "street_number_sorted": "123",
                    },
                    {
                        "is_pobox": False,
                        "street1": "1 1ST AVE",
                        "city": "SPRINGFIELD",
                        "state": None,
                        "postal_code": "62701",
                        "street_trigrams": [
                            "1 1",
                            " 1S",
                            "1ST",
                            "ST ",
                            "T A",
                            " AV",
                            "AVE",
                        ],
                        "street_name": "1ST",
                        "street_number": "1",
                        "street_number_sorted": "1",
                    },
                ],
                "addresses_keywords": [
                    " 1S",
                    " AV",
                    " MA",
                    " ST",
                    "1 1",
                    "132",
                    "1ST",
                    "2 M",
                    "32 ",
                    "AIN",
                    "AVE",
                    "IN ",
                    "MAI",
                    "N S",
                    "ST ",
                    "T A",
                ],
            },
            id="multi",
        ),
        pytest.param(
            [
                {
                    "street1": "132 Main St",
                    "street2": None,
                    "city": None,
                    "state": "IL IL",
                    "postal_code": "62701",
                    "country": "US",
                }
            ],
            {
                "addresses": [
                    {
                        "is_pobox": False,
                        "street1": "132 MAIN ST",
                        "city": None,
                        "state": "IL IL",
                        "postal_code": "62701",
                        "street_trigrams": [
                            "132",
                            "32 ",
                            "2 M",
                            " MA",
                            "MAI",
                            "AIN",
                            "IN ",
                            "N S",
                            " ST",
                        ],
                        "street_name": "MAIN",
                        "street_number": "132",
                        "street_number_sorted": "123",
                    }
                ],
                "addresses_keywords": [
                    "132",
                    "32 ",
                    "2 M",
                    " MA",
                    "MAI",
                    "AIN",
                    "IN ",
                    "N S",
                    " ST",
                ],
            },
            id="single-weird-state",
        ),
        pytest.param(
            [
                {
                    "street1": None,
                    "street2": None,
                    "city": None,
                    "state": None,
                    "postal_code": None,
                    "country": None,
                }
            ],
            {
                "addresses": [
                    {
                        "city": None,
                        "is_pobox": None,
                        "postal_code": None,
                        "state": None,
                        "street1": None,
                        "street_name": None,
                        "street_trigrams": None,
                        "street_number": None,
                        "street_number_sorted": None,
                    },
                ],
                "addresses_keywords": [],
            },
            id="all_null",
        ),
        pytest.param(
            [],
            {"addresses": [], "addresses_keywords": []},
            id="empty",
        ),
        pytest.param(None, {"addresses": None, "addresses_keywords": None}, id="null"),
    ],
)
def test_addresses_dimension(addresses, expected, table_factory):
    address_type = "array<struct<street1: string, street2: string, city: string, state: string, postal_code: string, country: string>>"  # noqa: E501
    t = table_factory({"addresses": [addresses]}, schema={"addresses": address_type})
    dim = geo.AddressesDimension("addresses")
    result = (
        dim.prepare_for_blocking(dim.prepare_for_fast_linking(t))
        .addresses_featured.execute()
        .iloc[0]
    )

    def setify(x):
        return set(x) if x is not None else None

    if result["addresses"] is not None:
        for address in result["addresses"]:
            address["street_trigrams"] = setify(address["street_trigrams"])
    if expected["addresses"] is not None:
        for address in expected["addresses"]:
            address["street_trigrams"] = setify(address["street_trigrams"])
    expected["addresses_keywords"] = setify(expected["addresses_keywords"])
    result["addresses_keywords"] = setify(result["addresses_keywords"])
    assert result == expected



from __future__ import annotations

from typing import Iterable

from mismo.linkage._linkage import Linkage


def _pair_counts_chart(linkages: Iterable[Linkage]):
    # Need to flesh out this API before I expose it
    import altair as alt

    counts = [
        {"blocker": linkage.name, "n": linkage.links.count()} for linkage in linkages
    ]
    # some of these counts are 0 (no pairs generated)
    # or None (no way to estimate)
    # At this point leave these in the chart.
    n_title = "Number of Pairs"
    chart = (
        alt.Chart(counts)
        .properties(
            title=alt.TitleParams(
                "Number of Pairs generated by each Blocker",
                subtitle=[
                    f"Total number of pairs: {counts.n.sum().execute():,}",
                    "If two blockers generate the same pair, then the pair will only appear once in the blocked output.",  # noqa: E501
                    "So, the sum of these counts is probably an overestimate of the true blocked size.",  # noqa: E501
                ],
                anchor="middle",
            ),
            height=150,
        )
        .mark_bar()
        .encode(
            alt.X("blocker:O", title="Blocker", sort="-y"),
            alt.Y("n:Q", title=n_title),
            tooltip=[
                alt.Tooltip("n:Q", title=n_title, format=","),
                alt.Tooltip("blocker:O", title="Blocker"),
            ],
        )
    )
    return chart



from __future__ import annotations

import logging
from typing import (
    Callable,
    Generic,
    Iterable,
    Protocol,
    TypeVar,
    runtime_checkable,
)

import ibis

from mismo._registry import Registry
from mismo.joins import HasJoinCondition, join
from mismo.linkage._linkage import Linkage, LinkedTable, LinksTable

logger = logging.getLogger(__name__)

L = TypeVar("L", bound=Linkage)


@runtime_checkable
class HasSublinkages(Protocol, Generic[L]):
    def __sublinkages__(self) -> tuple[L]:
        pass


class CombinedLinkage(Generic[L]):
    def __init__(self, sublinkages: Iterable[L]) -> None:
        self._sublinkages = tuple(sublinkages)
        if not self._sublinkages:
            raise ValueError("Need at least 1 linkage")

    def __sublinkages__(self) -> tuple[L]:
        return self._sublinkages

    @property
    def left(self) -> LinkedTable:
        return self._make_pair[0]

    @property
    def right(self) -> LinkedTable:
        return self._make_pair[1]

    @property
    def links(self) -> LinksTable:
        return LinksTable(self._links_raw, left=self._left_raw, right=self._left_raw)

    @property
    def _make_pair(self):
        return LinkedTable.make_pair(
            left=self._left_raw, right=self._right_raw, links=self._links_raw
        )

    @property
    def _left_raw(self) -> ibis.Table:
        return self._sublinkages[0].left

    @property
    def _right_raw(self) -> ibis.Table:
        return self._sublinkages[0].right

    @property
    def _links_raw(self) -> ibis.Table:
        raise NotImplementedError

    def cache(self) -> Linkage:
        """
        No-op to cache this, since the condition is an abstract condition, and
        left and right are already cached.
        """
        return self


class UnionLinkage(CombinedLinkage[L]):
    def __init__(self, sublinkages: Iterable[L]) -> None:
        flattened = []
        for sub in sublinkages:
            if isinstance(sub, UnionLinkage):
                flattened.extend(sub.__sublinkages__())
            else:
                flattened.append(sub)
        # TODO: I think we should be efficient here and do something like
        # union(
        #     linkages[0].links,
        #     <all the links in linkages[1] except add an additional join condition to
        #     not include links that are already in the first linkage>
        #     <all the links in linkages[2] except add an additional join condition to
        #     not include links that are already in the first or second linkage>
        #     <etc...>
        # )
        super().__init__(flattened)

    @property
    def _links_raw(self) -> ibis.Table:
        return ibis.union(*(sub.links for sub in self.__sublinkages__()))


class IntersectionLinkage(CombinedLinkage[L]):
    def __init__(self, sublinkages: Iterable[L]) -> None:
        flattened = []
        for sub in sublinkages:
            if isinstance(sub, IntersectionLinkage):
                flattened.extend(sub.__sublinkages__())
            else:
                flattened.append(sub)
        super().__init__(flattened)

    @property
    def _links_raw(self) -> ibis.Table:
        return ibis.intersect(*(sub.links for sub in self.__sublinkages__()))


class DifferenceLinkage(CombinedLinkage[L]):
    def __init__(self, sublinkages: Iterable[L]) -> None:
        # TODO: I think we can optimize this by flattening nested ones
        super().__init__(sublinkages)

    @property
    def _links_raw(self) -> ibis.Table:
        return ibis.difference(*(sub.links for sub in self.__sublinkages__()))


@runtime_checkable
class PCombiner(Protocol):
    def __call__(self, linkages: Iterable[Linkage]) -> Linkage: ...


class CombinerRegistry(Registry[PCombiner, Linkage]):
    def __call__(self, linkages: Iterable[Linkage]) -> Linkage:
        # Should we verify that all the linkages share the same left and right tables?
        linkages = tuple(linkages)
        non_linkages = [ln for ln in linkages if not isinstance(ln, Linkage)]
        if non_linkages:
            raise TypeError(f"received some non-Linkages: {non_linkages}", non_linkages)
        return super().__call__(linkages)


@runtime_checkable
class HasJoinConditionLinkage(HasJoinCondition, Protocol):
    pass


class AndJoinConditionsLinkage(CombinedLinkage[HasJoinConditionLinkage]):
    def __init__(self, sublinkages: Iterable[HasJoinConditionLinkage]):
        if not self.is_match(sublinkages):
            return NotImplemented
        flattened = []
        for sub in sublinkages:
            if isinstance(sub, AndJoinConditionsLinkage):
                flattened.extend(sub.__sublinkages__())
            else:
                flattened.append(sub)
        super().__init__(flattened)

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> Callable[[ibis.Table, ibis.Table], ibis.ir.BooleanValue]:
        return ibis.and_(
            *[sub.__join_condition__(left, right) for sub in self.__sublinkages__()]
        )

    @property
    def _links_raw(self) -> ibis.Table:
        return join(
            self._left_raw,
            self._right_raw,
            self.__join_condition__,
            lname="{name}_l",
            rname="{name}_r",
            rename_all=True,
        )

    @classmethod
    def is_match(cls, linkages: Iterable[Linkage]) -> bool:
        return all(isinstance(linkage, HasJoinConditionLinkage) for linkage in linkages)


union = CombinerRegistry()
"""Return a Linkage with links that are in any input Linkages."""
union.register(UnionLinkage)

intersect = CombinerRegistry()
"""Return a Linkage with links that are in all input Linkages."""
intersect.register(IntersectionLinkage)
intersect.register(AndJoinConditionsLinkage)

difference = CombinerRegistry()
"""
Return a Linkage with links that are in the first Linkage but not in any of the rest.
"""
difference.register(DifferenceLinkage)


def unify_links_min_intersection(*linkages: Linkage) -> tuple[Linkage, ...]:
    """
    Given some Linkages, simplify each links to only include the columns that are in all Linkages.

    In order to combine all the `LinksTable`s into a single `LinkTable`,
    we need all their schemas to be the same.
    This is one strategy to do that.

    Another strategy would be to include the columns that are in any of the link tables,
    and fill in the missing columns with nulls.
    """  # noqa: E501
    all_links = [linkage.links for linkage in linkages]
    shared, *rest = [
        set((name, dtype) for name, dtype in link.schema().items())
        for link in all_links
    ]
    for column_set in rest:
        shared &= column_set
    shared_columns = [name for name, _ in shared]
    return tuple(
        linkage.copy(links=linkage.links.select(*shared_columns))
        for linkage in linkages
    )



from __future__ import annotations

from collections.abc import Iterable
import typing

import ibis
from ibis import ir

from mismo.types._links_table import LinksTable


@typing.runtime_checkable
class Dimension(typing.Protocol):
    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ir.BooleanValue | bool:
        pass

    def __compare__(self, links: LinksTable) -> LinksTable:
        pass


class DimensionsLinker:
    def __init__(self, dimensions: Iterable[Dimension]) -> None:
        self.dimension = tuple(dimensions)

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ibis.ir.BooleanValue | bool:
        clauses = [linker.__join_condition__(left, right) for linker in self.dimension]
        clauses = [clause for clause in clauses if clause is not True]
        return ibis.and_(*clauses)

    def __compare__(self, links: LinksTable) -> LinksTable:
        for linker in self.dimension:
            links = linker.__compare__(links)
        return links



from __future__ import annotations

from itertools import count
import logging
from typing import Callable, Iterable, Mapping, overload

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo._datasets import Datasets
from mismo._factorizer import Factorizer

logger = logging.getLogger(__name__)


@overload
def connected_components(
    *,
    links: ir.Table,
    records: ir.Column | ir.Table | None,
    max_iter: int | None = None,
    label_as: str = "component",
) -> ir.Table: ...


@overload
def connected_components(
    *,
    links: ir.Table,
    records: Iterable[ir.Table] | Mapping[str, ir.Table],
    max_iter: int | None = None,
    label_as: str = "component",
) -> Datasets: ...


# I think more performant algorithms exist, but they are more complicated.
# See https://arxiv.org/pdf/1802.09478.pdf
def connected_components(
    *,
    links: ir.Table,
    records: ir.Column | ir.Table | Iterable[ir.Table] | Mapping[str, ir.Table] = None,
    max_iter: int | None = None,
    label_as: str = "component",
) -> ir.Table | Datasets:
    """Label records using connected components, based on the given links.

    This uses [an iterative algorithm](https://www.drmaciver.com/2008/11/computing-connected-graph-components-via-sql/)
    that is linear in terms of the diameter of the largest component
    (ie how many "hops" it takes to get from one end of a cluster to the other).
    This is usually acceptable for our use case,
    because we expect the components to be small.

    Parameters
    ----------
    links :
        A table with the columns (record_id_l, record_id_r), corresponding
        to the `record_id`s in `records`.
    records :
        Table(s) of records with at least the column `record_id`, the column
        of record_ids itself, or None.

        !!! note

            If you supply multiple Tables, the record_ids must be the same type
            across all tables, and **universally** unique across all tables

    max_iter :
        The maximum number of iterations to run. If None, run until convergence.

    label_as :
        The name of the label column that will contain the component ID.

    Returns
    -------
    result
        - If `records` is None, a Table will be returned with columns
        `record_id` and `<label_as>` of type `int64` that maps record_id to component.
        - If `records` is a single Table, that table will be returned
        with a `<label_as> column added of type `int64`.
        - If `records` is an iterable/mapping of Tables, a `Datasets` will be returned,
        with a `<label_as>` column of type `int64` added to each contained Table.

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> from mismo.cluster import connected_components
    >>> records1 = ibis.memtable(
    ...     [
    ...         ("a", 0),
    ...         ("b", 1),
    ...         ("c", 2),
    ...         ("d", 3),
    ...         ("g", 6),
    ...     ],
    ...     columns=["record_id", "other"],
    ... )
    >>> records2 = ibis.memtable(
    ...     [
    ...         ("h", 7),
    ...         ("x", 23),
    ...         ("y", 24),
    ...         ("z", 25),
    ...     ],
    ...     columns=["record_id", "other"],
    ... )
    >>> links = ibis.memtable(
    ...     [
    ...         ("a", "x"),
    ...         ("b", "x"),
    ...         ("b", "y"),
    ...         ("c", "y"),
    ...         ("c", "z"),
    ...         ("g", "h"),
    ...     ],
    ...     columns=["record_id_l", "record_id_r"],
    ... )

    If you don't supply the records, then you just get a labeling map
    from record_id -> component. Note how only the record_ids that are
    present in `links` are returned, eg there is no record_id `"d"` present:

    >>> connected_components(links=links).order_by("record_id")
    ┏━━━━━━━━━━━┳━━━━━━━━━━━┓
    ┃ record_id ┃ component ┃
    ┡━━━━━━━━━━━╇━━━━━━━━━━━┩
    │ string    │ int64     │
    ├───────────┼───────────┤
    │ a         │         0 │
    │ b         │         0 │
    │ c         │         0 │
    │ g         │         3 │
    │ h         │         3 │
    │ x         │         0 │
    │ y         │         0 │
    │ z         │         0 │
    └───────────┴───────────┘

    If you supply records, then the records are labeled with the component.
    We can also change the name of the column that contains the component:

    >>> connected_components(
    ...     records=records1,
    ...     links=links,
    ...     label_as="label",
    ... ).order_by("record_id")
    ┏━━━━━━━━━━━┳━━━━━━━┳━━━━━━━┓
    ┃ record_id ┃ other ┃ label ┃
    ┡━━━━━━━━━━━╇━━━━━━━╇━━━━━━━┩
    │ string    │ int64 │ int64 │
    ├───────────┼───────┼───────┤
    │ a         │     0 │     0 │
    │ b         │     1 │     0 │
    │ c         │     2 │     0 │
    │ d         │     3 │     4 │
    │ g         │     6 │     3 │
    └───────────┴───────┴───────┘

    You can supply multiple sets of records, which are coerced to a `Datasets`,
    and returned as a `Datasets`, with each table of records labeled
    individually.

    >>> a, b = connected_components(records=(records1, records2), links=links)
    >>> a.order_by("record_id")
    ┏━━━━━━━━━━━┳━━━━━━━┳━━━━━━━━━━━┓
    ┃ record_id ┃ other ┃ component ┃
    ┡━━━━━━━━━━━╇━━━━━━━╇━━━━━━━━━━━┩
    │ string    │ int64 │ int64     │
    ├───────────┼───────┼───────────┤
    │ a         │     0 │         0 │
    │ b         │     1 │         0 │
    │ c         │     2 │         0 │
    │ d         │     3 │         4 │
    │ g         │     6 │         3 │
    └───────────┴───────┴───────────┘
    >>> b.order_by("record_id")
    ┏━━━━━━━━━━━┳━━━━━━━┳━━━━━━━━━━━┓
    ┃ record_id ┃ other ┃ component ┃
    ┡━━━━━━━━━━━╇━━━━━━━╇━━━━━━━━━━━┩
    │ string    │ int64 │ int64     │
    ├───────────┼───────┼───────────┤
    │ h         │     7 │         3 │
    │ x         │    23 │         0 │
    │ y         │    24 │         0 │
    │ z         │    25 │         0 │
    └───────────┴───────┴───────────┘
    """  # noqa: E501
    int_edges, restore = _intify_edges(links)
    int_labels = _connected_components_ints(int_edges, max_iter=max_iter)
    labels = restore(int_labels)
    if records is None:
        return labels.rename(**{label_as: "component"})
    if isinstance(records, ir.Column):
        records = records.name("record_id").as_table()
    ds = Datasets(records)
    result = _label_datasets(ds, labels, label_as=label_as)
    if isinstance(records, ir.Table):
        return result[0]
    return result


def _connected_components_ints(
    edges: ir.Table, max_iter: int | None = None
) -> tuple[ir.Table, ir.Table]:
    """The core algorithm. Assumes you already translated the record ids to ints."""
    labels = _get_initial_labels(edges).cache()
    for i in count(1):
        new_labels = _updated_labels(labels, edges).cache()
        assert labels.count().execute() == new_labels.count().execute()
        n_updates = _n_updates(labels, new_labels)
        if n_updates == 0:
            return labels
        logger.info(f"Round {i}: Updated {n_updates} labels")
        labels = new_labels
        if max_iter is not None and i >= max_iter:
            return labels


def _n_updates(labels: ir.Table, new_labels: ir.Table) -> int:
    """Count the number of updates between two labelings."""
    condition = (labels.record_id == new_labels.record_id) & (
        labels.component != new_labels.component
    )
    return ibis.join(labels, new_labels, condition).count().execute()


def _updated_labels(node_labels: ir.Table, edges: ir.Table) -> ir.Table:
    component_equivalences = _get_component_equivalences(edges, node_labels)
    component_mapping = _get_component_update_map(component_equivalences)
    return node_labels.rename(component_old="component").left_join(
        component_mapping, "component_old"
    )["record_id", "component"]


def _get_component_equivalences(edges: ir.Table, node_labels: ir.Table) -> ir.Table:
    """Get a table of which components are equivalent to each other."""
    same_components = (
        edges.join(node_labels, edges.record_id_l == node_labels.record_id)
        .rename(component_l="component")
        .drop("record_id", "record_id_l")
    )
    same_components = (
        same_components.join(
            node_labels, same_components.record_id_r == node_labels.record_id
        )
        .rename(component_r="component")
        .drop("record_id", "record_id_r")
    )
    return same_components["component_l", "component_r"].distinct()


def _get_component_update_map(component_equivalences: ir.Table) -> ir.Table:
    """Create mapping from old component ids to new component ids"""
    representative = ibis.least(
        component_equivalences.component_l, component_equivalences.component_r
    )
    m = component_equivalences.mutate(component=representative)
    ml = m.select("component", component_old="component_l")
    mr = m.select("component", component_old="component_r")
    together = ibis.union(ml, mr, distinct=True)
    together = together.group_by(_.component_old).agg(component=_.component.min())
    return together["component_old", "component"]


def _intify_edges(
    raw_edges: ir.Table,
) -> tuple[ir.Table, Callable[[ir.Table], ir.Table]]:
    """Translate edges to int64s and create restoring function"""
    if "record_id_l" not in raw_edges.columns or "record_id_r" not in raw_edges.columns:
        raise ValueError(
            "edges must contain the columns `record_id_l` and `record_id_r`, "
            f"but it contains {raw_edges.columns}"
        )
    raw_edges = raw_edges.select("record_id_l", "record_id_r")
    swapped = raw_edges.select(record_id_r=_.record_id_l, record_id_l=_.record_id_r)
    all_node_ids = raw_edges.union(swapped).select("record_id_l")

    f = Factorizer(all_node_ids, "record_id_l")
    edges = f.encode(raw_edges, "record_id_l")
    edges = f.encode(edges, "record_id_r")

    def restore(int_labels: ir.Table) -> ir.Table:
        return f.decode(int_labels, "record_id")

    return edges, restore


def _get_initial_labels(edges: ir.Table) -> ir.Table:
    labels_left = edges.select(record_id=_.record_id_l, component=_.record_id_l)
    labels_right = edges.select(record_id=_.record_id_r, component=_.record_id_r)
    return ibis.union(labels_left, labels_right, distinct=True)


def _label_datasets(ds: Datasets, labels: ir.Table, *, label_as: str) -> Datasets:
    additional_labels = _get_additional_labels(labels, ds.all_record_ids())
    labels = labels.union(additional_labels)
    return ds.map(
        lambda name, t: t.left_join(labels, "record_id", rname="{name}_ibis_tmp")
        .drop("record_id_ibis_tmp")
        .rename(**{label_as: "component"})
    )


def _get_additional_labels(labels: ir.Table, record_ids: ir.Column) -> ir.Table:
    nodes = record_ids.name("record_id").as_table()
    is_missing_label = nodes.record_id.notin(labels.record_id)
    # fill_null(0) in case labels is empty
    max_existing_label = labels.component.max().fill_null(0)
    additional_labels = nodes.filter(is_missing_label).select(
        "record_id",
        component=(ibis.row_number() + max_existing_label + 1).cast("int64"),
    )
    return additional_labels



from __future__ import annotations

import pytest

from mismo.linkage import sample_all_links


@pytest.mark.parametrize(
    "n_records,max_pairs",
    [
        (10_000, 0),
        (10_000, 1),
        (10_000, 2),
        (10_000, 10),
        (10_000, 100_000),
        # Stress test the implementation we
        # use to ensure that pairs are not duplicated.
        (2, 3),
        (10, 99),
        # Edge case where n_possible_pairs < max_pairs
        (3, 10_000_000),
    ],
)
def test_sample_all_pairs(table_factory, n_records: int, max_pairs: int):
    t = table_factory(
        {
            "record_id": range(n_records),
            # Something to check that the column is not included in output
            "value": [i // 7 for i in range(n_records)],
        }
    )
    df = sample_all_links(t, t, max_pairs=max_pairs).execute()
    expected_pairs = min(n_records**2, max_pairs)
    assert df.columns.tolist() == ["record_id_l", "record_id_r"]
    assert len(df) == expected_pairs
    assert df.record_id_l.notnull().all()
    assert df.record_id_r.notnull().all()
    # We get no duplicates
    assert len(df.drop_duplicates()) == expected_pairs

    # Only do these stats if we have a reasonable number of pairs
    if expected_pairs >= 1000:
        # These two aren't correlated
        assert (df.record_id_l == df.record_id_r).mean() < 0.01
        # We get about an even distribution, the head or tail are not overrepresented
        # We expect the mean of record ids to be about half of the n_pairs
        expected = n_records / 2
        pct_err_l = abs(df.record_id_l.mean() - expected) / expected
        pct_err_r = abs(df.record_id_r.mean() - expected) / expected
        assert pct_err_l < 0.01
        assert pct_err_r < 0.01


@pytest.mark.parametrize("max_pairs", [None, 0, 1, 100_000])
def test_sample_all_pairs_empty(table_factory, max_pairs: int | None):
    t = table_factory({"record_id": []})
    df = sample_all_links(t, t, max_pairs=max_pairs).execute()
    assert df.columns.tolist() == ["record_id_l", "record_id_r"]
    assert len(df) == 0


def test_sample_all_pairs_warns(table_factory):
    t = table_factory({"record_id": range(100_000)})
    with pytest.warns(UserWarning):
        sample_all_links(t, t)


def test_sample_all_pairs_different_tables(table_factory):
    # check that we can sample from two different tables
    # https://github.com/NickCrews/mismo/issues/35#issuecomment-2116178262
    t = table_factory({"record_id": range(100)})
    df = sample_all_links(t, t.view(), max_pairs=10).execute()
    assert df.columns.tolist() == ["record_id_l", "record_id_r"]
    assert len(df) == 10



from __future__ import annotations

import ibis
from ibis import _
from ibis.expr import datatypes as dt
from ibis.expr import types as ir

from mismo import _util, vector


def document_counts(terms: ir.ArrayColumn) -> ir.Table:
    r"""Create a lookup Table from term to number of records containing the term.

    Parameters
    ----------
    terms:
        One row for each record. Each row is an array of terms in that record.
        Each term could be a word, ngram, or other token from a string.
        Or, it could also represent more generic data, such as a list of tags or
        categories like ["red", "black"]. Each term can be any datatype,
        not just strings.

    Returns
    -------
        A Table with columns `term` and `n_records`. The `term` column contains
        each unique term from the input `terms` array. The `n_records` column
        contains the number of records in the input `terms` array that contain

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> from mismo.sets import document_counts
    >>> addresses = [
    ...     "12 main st",
    ...     "99 main ave",
    ...     "56 st joseph st",
    ...     "21 glacier st",
    ...     "12 glacier st",
    ... ]
    >>> t = ibis.memtable({"address": addresses})
    >>> # split on whitespace
    >>> t = t.mutate(terms=t.address.re_split(r"\s+"))
    >>> document_counts(t.terms).order_by("term")
    ┏━━━━━━━━━┳━━━━━━━━━━━┓
    ┃ term    ┃ n_records ┃
    ┡━━━━━━━━━╇━━━━━━━━━━━┩
    │ string  │ int64     │
    ├─────────┼───────────┤
    │ 12      │         2 │
    │ 21      │         1 │
    │ 56      │         1 │
    │ 99      │         1 │
    │ ave     │         1 │
    │ glacier │         2 │
    │ joseph  │         1 │
    │ main    │         2 │
    │ st      │         4 │
    └─────────┴───────────┘
    """  # noqa: E501
    if not isinstance(terms, ir.ArrayValue):
        raise ValueError(f"Unsupported type {type(terms)}")
    terms_table = (
        terms.name("terms")
        .as_table()
        .select(
            record_id=ibis.row_number(),
            terms=_.terms,
        )
    )
    flat = terms_table.select("record_id", term=_.terms.unnest())
    by_term = flat.group_by("term").agg(n_records=_.record_id.nunique())
    return by_term


def term_idf(terms: ir.ArrayValue) -> ir.Table:
    r"""Create a lookup Table from term to IDF.

    Examples
    --------
    >>> import ibis
    >>> from mismo.sets import term_idf
    >>> ibis.options.interactive = True
    >>> addresses = [
    ...     "12 main st",
    ...     "99 main ave",
    ...     "56 st joseph st",
    ...     "21 glacier st",
    ...     "12 glacier st",
    ... ]
    >>> t = ibis.memtable({"address": addresses})
    >>> # split on whitespace
    >>> t = t.mutate(terms=t.address.re_split(r"\s+"))
    >>> term_idf(t.terms).order_by("term")
    ┏━━━━━━━━━┳━━━━━━━━━━┓
    ┃ term    ┃ idf      ┃
    ┡━━━━━━━━━╇━━━━━━━━━━┩
    │ string  │ float64  │
    ├─────────┼──────────┤
    │ 12      │ 0.916291 │
    │ 21      │ 1.609438 │
    │ 56      │ 1.609438 │
    │ 99      │ 1.609438 │
    │ ave     │ 1.609438 │
    │ glacier │ 0.916291 │
    │ joseph  │ 1.609438 │
    │ main    │ 0.916291 │
    │ st      │ 0.223144 │
    └─────────┴──────────┘
    """
    dc = document_counts(terms)
    n_total_records = terms.count()
    idf = dc.select(
        "term",
        idf=(n_total_records / dc.n_records).log(),
    )
    return idf


# TODO: if https://github.com/ibis-project/ibis/issues/8614
# is fixed, this API can improve to be (ArrayValue) -> MapValue
def add_array_value_counts(
    t: ir.Table, column: str, *, result_name: str = "{name}_counts"
) -> ir.Table:
    r"""value_counts() for ArrayColumns.

    Parameters
    ----------
    t : Table
        The input table.
    column : str
        The name of the array column to analyze.
    result_name : str, optional
        The name of the resulting column. The default is "{name}_counts".

    Examples
    --------
    >>> import ibis
    >>> from mismo.sets import add_array_value_counts
    >>> ibis.options.interactive = True
    >>> ibis.options.repr.interactive.max_length = 20
    >>> terms = [
    ...     None,
    ...     ["st"],
    ...     ["st"],
    ...     ["12", "main", "st"],
    ...     ["99", "main", "ave"],
    ...     ["56", "st", "joseph", "st"],
    ...     ["21", "glacier", "st"],
    ...     ["12", "glacier", "st"],
    ... ]
    >>> t = ibis.memtable({"terms": terms})
    >>> add_array_value_counts(t, "terms")  # doctest: +SKIP
    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ terms                        ┃ terms_counts                     ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
    │ array<string>                │ map<string, int64>               │
    ├──────────────────────────────┼──────────────────────────────────┤
    │ ['st']                       │ {'st': 1}                        │
    │ ['st']                       │ {'st': 1}                        │
    │ ['12', 'main', 'st']         │ {'st': 1, '12': 1, 'main': 1}    │
    │ ['99', 'main', 'ave']        │ {'ave': 1, '99': 1, 'main': 1}   │
    │ ['56', 'st', 'joseph', 'st'] │ {'56': 1, 'joseph': 1, 'st': 2}  │
    │ ['21', 'glacier', 'st']      │ {'glacier': 1, 'st': 1, '21': 1} │
    │ ['12', 'glacier', 'st']      │ {'glacier': 1, 'st': 1, '12': 1} │
    │ NULL                         │ NULL                             │
    └──────────────────────────────┴──────────────────────────────────┘
    """  # noqa: E501
    t = t.mutate(__terms=_util.get_column(t, column))
    normalized = (
        t.select("__terms")
        .distinct()
        .mutate(
            __term=_.__terms.filter(lambda t: t.notnull()).unnest(),
        )
    )
    counts = normalized.group_by(["__terms", "__term"]).agg(
        __n=_.count(),
    )
    by_terms = counts.group_by("__terms").agg(
        __result=ibis.map(_.__term.collect(), _.__n.collect()),
    )
    result = t.left_join(by_terms, "__terms").drop("__terms", "__terms_right")

    # annoying logic to deal with the edge case of an empty array
    term_type = t.__terms.type().value_type
    counts_type = dt.Map(key_type=term_type, value_type=dt.int64)
    empty = ibis.literal({}, counts_type)
    result = result.mutate(__result=(_[column].length() == 0).ifelse(empty, _.__result))
    return result.rename({result_name.format(name=column): "__result"})


def add_tfidf(
    t,
    column: str,
    *,
    result_name: str = "{name}_tfidf",
    normalize: bool = True,
):
    r"""Vectorize terms using TF-IDF.

    Adds a column to the input table that contains the TF-IDF vector for the
    terms in the input column.

    Parameters
    ----------
    t : Table
        The input table.
    column : str
        The name of the array column to analyze.
    result_name : str, optional
        The name of the resulting column. The default is "{name}_tfidf".
    normalize :
        Whether to normalize the TF-vector before multiplying by the IDF.
        The default is True.
        This makes it so that vectors of different lengths can be compared fairly.

    Examples
    --------
    >>> import ibis
    >>> from mismo.sets import add_tfidf
    >>> ibis.options.interactive = True
    >>> ibis.options.repr.interactive.max_length = 20
    >>> terms = [
    ...     None,
    ...     ["st"],
    ...     ["st"],
    ...     ["12", "main", "st"],
    ...     ["99", "main", "ave"],
    ...     ["56", "st", "joseph", "st"],
    ...     ["21", "glacier", "st"],
    ...     ["12", "glacier", "st"],
    ... ]
    >>> t = ibis.memtable({"terms": terms})
    >>> add_tfidf(t, "terms")  # doctest: +SKIP
    ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ terms                        ┃ terms_tfidf                                                                          ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
    │ array<string>                │ map<string, float64>                                                                 │
    ├──────────────────────────────┼──────────────────────────────────────────────────────────────────────────────────────┤
    │ ['st']                       │ {'st': 0.15415067982725836}                                                          │
    │ ['st']                       │ {'st': 0.15415067982725836}                                                          │
    │ ['12', 'glacier', 'st']      │ {'12': 0.7232830370915955, 'glacier': 0.7232830370915955, 'st': 0.08899893649403144} │
    │ ['12', 'main', 'st']         │ {'12': 0.7232830370915955, 'main': 0.7232830370915955, 'st': 0.08899893649403144}    │
    │ ['21', 'glacier', 'st']      │ {'21': 1.12347174837591, 'glacier': 0.7232830370915955, 'st': 0.08899893649403144}   │
    │ ['56', 'st', 'joseph', 'st'] │ {'56': 0.7944144917481126, 'joseph': 0.7944144917481126, 'st': 0.12586350302664107}  │
    │ ['99', 'main', 'ave']        │ {'main': 0.7232830370915955, 'ave': 1.12347174837591, '99': 1.12347174837591}        │
    │ NULL                         │ NULL                                                                                 │
    └──────────────────────────────┴──────────────────────────────────────────────────────────────────────────────────────┘
    """  # noqa: E501
    with_counts = add_array_value_counts(t, column, result_name="__term_counts")
    if normalize:
        with_counts = with_counts.mutate(
            __term_counts=vector.normalize(with_counts.__term_counts)
        )
    idf = term_idf(t[column])
    with_counts = with_counts.mutate(__row_number=ibis.row_number())
    term_counts = with_counts.select("__row_number", "__term_counts")
    term_counts = term_counts.mutate(
        term=_.__term_counts.keys().unnest(),
        term_count=_.__term_counts.values().unnest(),
    ).drop("__term_counts")
    idf_table = term_counts.join(idf, "term")
    idf_table = idf_table.mutate(idf=_.idf * _.term_count).drop("term_count")
    r = result_name.format(name=column)
    idf_table = idf_table.group_by("__row_number").aggregate(
        ibis.map(idf_table.term.collect(), idf_table.idf.collect()).name(r)
    )
    result = with_counts.left_join(idf_table, "__row_number").drop(
        "__row_number", "__term_counts", "__row_number_right"
    )
    empty = ibis.literal({}, type=result[r].type())
    result = result.mutate((_[column].length() == 0).ifelse(empty, _[r]).name(r))
    return result


def rare_terms(
    terms: ir.ArrayColumn,
    *,
    max_records_n: int | None = None,
    max_records_frac: float | None = None,
) -> ir.Column:
    """Get the terms that appear in few records.

    The returned Column is flattened. Eg if you supply a column of `array<string>`,
    the result will be of type `string`.

    Exactly one of `max_records_n` or `max_records_frac` must be set.

    Parameters
    ----------
    terms : ArrayColumn
        A column of Arrays, where each array contains the terms for a record.
    max_records_n : int, optional
        The maximum number of records a term can appear in. The default is None.
    max_records_frac : float, optional
        The maximum fraction of records a term can appear in. The default is None.

    Returns
    -------
    Column
        The terms that appear in few records.
    """
    if max_records_n is not None and max_records_frac is not None:
        raise ValueError("Only one of max_records_n or max_records_frac can be set")
    if max_records_n is None and max_records_frac is None:
        raise ValueError("One of max_records_n or max_records_frac must be set")
    dc = document_counts(terms)
    if max_records_n is not None:
        rare = dc.filter(_.n_records <= max_records_n)
    else:
        n_total_records = terms.count()
        dc = dc.mutate(frac=_.n_records / n_total_records)
        rare = dc.filter(_.frac <= max_records_frac)
    return rare["term"]



from __future__ import annotations

import typing

import ibis

from mismo import _util

if typing.TYPE_CHECKING:
    from mismo.joins._analyze import SlowJoinAlgorithm


class MismoError(Exception):
    """Base class for all Mismo errors."""


class MismoWarning(Warning):
    """Base class for all Mismo warnings."""


class UnsupportedBackendError(ValueError, MismoError):
    """An operation is not supported by a particular backend."""


class SlowJoinMixin:
    def __init__(
        self, condition: ibis.ir.BooleanValue, algorithm: SlowJoinAlgorithm
    ) -> None:
        self.condition: ibis.ir.BooleanValue = condition
        """The join condition that is slow."""
        self.algorithm: SlowJoinAlgorithm = algorithm
        """The algorithm used for the join, eg 'NESTED_LOOP_JOIN'."""
        super().__init__(
            f"The join '{_util.get_name(self.condition)}' uses the {algorithm} algorithm and is likely to be slow."  # noqa: E501
        )


class SlowJoinWarning(SlowJoinMixin, UserWarning, MismoWarning):
    """Warning for slow join algorithms."""


class SlowJoinError(SlowJoinMixin, ValueError, MismoError):
    """Error for slow join algorithms."""



from __future__ import annotations

from collections import defaultdict
import re

import ibis
from ibis.expr import datatypes as dt
from ibis.expr import types as ir

from mismo import _util

_ADDRESS_SCHEMA = dt.Struct(
    {
        "street1": "string",
        "street2": "string",
        "city": "string",
        "state": "string",
        "postal_code": "string",
        "country": "string",
    }
)

_DIGITS_REGEX = re.compile(r"[0-9]+")


def postal_parse_address(address_string: ir.StringValue) -> ir.StructValue:
    """Parse individual fields from an address string.

    .. note:: To use this function, you need the optional `postal` library installed.

    This uses the optional `postal` library to extract individual fields
    from the string using the following mapping:

    - house_number + road -> street1
    - unit -> street2
    - city -> city
    - state -> state
    - postcode -> postal_code
    - country -> country

    Any additional fields parsed by postal will not be included.

    Parameters
    ----------
    address_string :
        The address as a single string

    Returns
    -------
    address :
        The parsed address as a Struct
    """
    with _util.optional_import("postal"):
        from postal.parser import parse_address as _parse_address

    @ibis.udf.scalar.python(signature=((str,), _ADDRESS_SCHEMA))
    def udf(address_string: str | None) -> dict[str, str] | None:
        # remove once https://github.com/ibis-project/ibis/pull/9625 is fixed
        if address_string is None:
            return None
        parsed_fields = _parse_address(address_string)
        label_to_values = defaultdict(list)
        for value, label in parsed_fields:
            label_to_values[label].append(value)
        renamed = {
            "street1": label_to_values["house_number"] + label_to_values["road"],
            "street2": label_to_values["unit"],
            "city": label_to_values["city"],
            "state": label_to_values["state"],
            "postal_code": label_to_values["postcode"],
            "country": label_to_values["country"],
        }
        # replace empty strings with None
        return {k: " ".join(v) or None for k, v in renamed.items()}

    return udf(address_string)


def postal_fingerprint_address(address: ir.StructValue) -> ir.ArrayValue:
    """Generate multiple hashes of an address string to be used for e.g. blocking.

    .. note:: To use this function, you need to have the optional `postal` library
    installed.

    This uses the near-dupe hashing functionality of `postal` to expand the root
    of each address component, ignoring tokens such as "road" or "street" in
    street names.

    For street names, whitespace is removed so that for example "Sea Grape Ln" and
    "Seagrape Ln" will both normalize to "seagrape".

    This returns a list of normalized tokens that are the minimum
    required information to represent the given address.

    Near-dupe hashes can be used as keys when blocking,
    to generate pairs of potential duplicates.

    Further details about the hashing function can be found
    [here](https://github.com/openvenues/libpostal/releases/tag/v1.1).

    Note that `postal.near_dupe.near_dupe_hashes` can optionally hash names and
    use latlon coordinates for geohashing, but this function only hashes addresses.
    Name and geo-hashing must be implemented elsewhere

    Examples
    -------
    >>> address = ibis.struct(
    ...     {
    ...         "street1": "123 Main Street",
    ...         "street2": "",
    ...         "city": "Springfield",
    ...         "state": "IL",
    ...         "postal_code": "62701",
    ...         "country": "us",
    ...     }
    ... )
    >>> postal_fingerprint_address(address).execute()  # doctest: +SKIP
    [
        "act|main street|123|springfield",
        "act|main|123|springfield",
        "apc|main street|123|62701",
        "apc|main|123|62701",
    ]

    Parameters
    ----------
    address :
        The address

    Returns
    -------
    address_hashes :
        Hashes of the address.
    """
    with _util.optional_import("postal"):
        from postal.near_dupe import near_dupe_hashes as _hash

    @ibis.udf.scalar.python(signature=((_ADDRESS_SCHEMA,), list[str]))
    def udf(address: dict[str, str] | None) -> list[str] | None:
        # remove once https://github.com/ibis-project/ibis/pull/9625 is fixed
        if address is None:
            return None
        # split street1 into house_number and road
        street1 = address["street1"] or ""
        house, *rest = street1.split(" ", 1)
        contains_digits = _DIGITS_REGEX.match(house) is not None
        parsed = {
            "unit": address["street2"],
            "city": address["city"],
            "state": address["state"],
            "postcode": address["postal_code"],
            "country": address["country"],
        }
        if contains_digits:
            # handle the fact that street1 contains both the house number and the road
            parsed["house_number"] = house
            parsed["road"] = " ".join(rest)
        else:
            parsed["road"] = street1

        parsed = {k: v for k, v in parsed.items() if v}

        if len(parsed) == 0:
            # catch empty strings from invalid addresses
            return []
        return _hash(
            list(parsed.keys()),
            list(parsed.values()),
            address_only_keys=True,
        )

    return udf(address)



from __future__ import annotations

import ibis
from ibis.expr import types as ir
import pytest

import mismo
from mismo import JoinLinker
from mismo.tests.util import assert_tables_equal


def joined_on_letter(t1, t2, **kwargs):
    return ibis.join(
        t1, t2, t1.letter == t2.letter, lname="{name}_l", rname="{name}_r"
    )["record_id_l", "record_id_r"]


def joined_on_letter_int(t1, t2, **kwargs):
    return ibis.join(
        t1,
        t2,
        (t1.letter == t2.letter) & (t1.int == t2.int),
        lname="{name}_l",
        rname="{name}_r",
    )["record_id_l", "record_id_r"]


def joined_on_int_gt(t1, t2, **kwargs):
    return ibis.join(t1, t2, t1.int > t2.int, lname="{name}_l", rname="{name}_r")[
        "record_id_l", "record_id_r"
    ]


def letter_condition(left, right):
    return left.letter == right.letter


def letter_and_int_condition(left, right):
    return (left.letter == right.letter) & (left.int == right.int)


def int_gt_condition(left, right):
    return left.int > right.int


def cross_product_condition(left, right):
    return left.int > 0


def null_safe_letter_condition(left, right):
    return (left.letter == right.letter) | (
        left.letter.isnull() & right.letter.isnull()
    )


def name_condition(left, right):
    return left.name == right.full_name


def street_condition(left, right):
    return left.street == right.street


@pytest.mark.parametrize(
    "condition,expected_maker",
    [
        pytest.param(
            letter_condition,
            joined_on_letter,
            id="letter_equality",
        ),
        pytest.param(
            letter_and_int_condition,
            joined_on_letter_int,
            id="letter_and_int_equality",
        ),
        pytest.param(
            int_gt_condition,
            joined_on_int_gt,
            id="int_greater_than",
        ),
        # Test with deferred expressions
        pytest.param(
            letter_condition,
            joined_on_letter,
            id="deferred_letter",
        ),
    ],
)
def test_JoinLinker_conditions(t1: ir.Table, t2: ir.Table, condition, expected_maker):
    # Use on_slow="ignore" for tests that might have slow joins
    on_slow = "ignore" if condition == int_gt_condition else "error"
    linkage = JoinLinker(condition, on_slow=on_slow)(t1, t2)
    joined_ids = linkage.links.select("record_id_l", "record_id_r")
    expected = expected_maker(t1, t2)
    assert_tables_equal(joined_ids, expected)


def test_deduplication(table_factory):
    t = table_factory(
        {
            "record_id": [0, 1, 2, 3, 4],
            "letter": ["a", "b", "a", "b", None],
            "int": [1, 2, 1, 2, 3],
        }
    )
    linker = JoinLinker(letter_condition)
    linkage = linker(t, t)
    joined_ids = linkage.links.select("record_id_l", "record_id_r")

    # JoinLinker includes ALL matches (including self-matches) but applies
    # deduplication filter (record_id_l < record_id_r) only to self-joins
    # So we get: (0,2) for "a" matches, (1,3) for "b" matches
    expected = table_factory(
        {
            "record_id_l": [0, 1],
            "record_id_r": [2, 3],
        }
    )
    assert_tables_equal(joined_ids, expected)


def test_task_parameter(table_factory):
    t1 = table_factory(
        {
            "record_id": [0, 1],
            "letter": ["a", "b"],
        }
    )
    t2 = table_factory(
        {
            "record_id": [90, 91],
            "letter": ["a", "b"],
        }
    )

    linker = JoinLinker(letter_condition, task="link")
    linkage = linker(t1, t2)
    joined_ids = linkage.links.select("record_id_l", "record_id_r")
    expected = table_factory(
        {
            "record_id_l": [0, 1],
            "record_id_r": [90, 91],
        }
    )
    assert_tables_equal(joined_ids, expected)


def test_on_slow_parameter(t1: ir.Table, t2: ir.Table):
    # Should not raise with on_slow="ignore"
    linker = JoinLinker(cross_product_condition, on_slow="ignore")
    linkage = linker(t1, t2)
    assert linkage.links.count().execute() > 0

    # Should warn with on_slow="warn" (but not fail)
    import warnings

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        linker = JoinLinker(cross_product_condition, on_slow="warn")
        linkage = linker(t1, t2)
        assert linkage.links.count().execute() > 0
        # Should have issued a warning
        assert len(w) > 0
        assert "slow" in str(w[0].message).lower()


def test_complex_conditions(t1: ir.Table, t2: ir.Table):
    linker = JoinLinker(null_safe_letter_condition)
    linkage = linker(t1, t2)
    joined_ids = linkage.links.select("record_id_l", "record_id_r")

    # Should include the null-null matches
    assert joined_ids.count().execute() >= 2  # At least b-b and c-c matches


def test_repr():
    linker = JoinLinker(letter_condition, task="link", on_slow="warn")
    repr_str = repr(linker)
    assert "JoinLinker" in repr_str
    assert "task=link" in repr_str
    assert "on_slow=warn" in repr_str


def test_empty_result(table_factory):
    t1 = table_factory(
        {
            "record_id": [0, 1],
            "letter": ["a", "b"],
        }
    )
    t2 = table_factory(
        {
            "record_id": [90, 91],
            "letter": ["x", "y"],
        }
    )

    linker = JoinLinker(letter_condition)
    linkage = linker(t1, t2)
    joined_ids = linkage.links.select("record_id_l", "record_id_r")
    assert joined_ids.count().execute() == 0


@pytest.mark.parametrize(
    "fn",
    [
        pytest.param(
            lambda t: t.inner_join(t.view(), "street").count(),
            id="naive",
        ),
        pytest.param(
            lambda t: mismo.JoinLinker(street_condition)
            .pair_counts(t, t, task="link")
            .n.sum(),
            id="JoinLinker",
        ),
    ],
)
@pytest.mark.parametrize(
    "nrows,exp",
    [
        (1_000, 16_990),
        (10_000, 576_101),
        (100_000, 50_843_827),
        (300_000, 447_872_405),
    ],
)
def test_benchmark_n_pairs(addresses_1M, fn, nrows, exp, benchmark):
    """Benchmark test similar to KeyLinker but using JoinLinker."""

    def run():
        return fn(addresses_1M.head(nrows)).execute()

    result = benchmark(run)
    assert result == exp


def test_join_condition_with_different_column_names(table_factory):
    t1 = table_factory(
        {
            "record_id": [0, 1],
            "name": ["Alice", "Bob"],
        }
    )
    t2 = table_factory(
        {
            "record_id": [90, 91],
            "full_name": ["Alice", "Charlie"],
        }
    )

    linker = JoinLinker(name_condition)
    linkage = linker(t1, t2)
    joined_ids = linkage.links.select("record_id_l", "record_id_r")

    expected = table_factory(
        {
            "record_id_l": [0],
            "record_id_r": [90],
        }
    )
    assert_tables_equal(joined_ids, expected)



from __future__ import annotations

from typing import Literal

import ibis
from ibis.expr import types as ir
import pandas as pd
import pytest


def assert_tables_equal(
    left: ir.Table,
    right: ir.Table,
    *,
    column_order: Literal["ignore", "exact"] = "exact",
    on_schema_mismatch: Literal["error", "cast_to_left", "cast_to_right"] = "error",
    order_by=None,
) -> None:
    if dict(left.schema()) != dict(right.schema()):
        if on_schema_mismatch == "error":
            assert dict(left.schema()) == dict(right.schema())
        if on_schema_mismatch == "cast_to_left":
            right = right.cast(left.schema())
        elif on_schema_mismatch == "cast_to_right":
            left = left.cast(right.schema())
        else:
            raise ValueError(f"Invalid on_schema_mismatch: {on_schema_mismatch}")
    assert dict(left.schema()) == dict(right.schema())
    if column_order == "exact":
        assert left.columns == right.columns
    elif column_order == "ignore":
        left = left[left.columns]
        right = right[left.columns]
    else:
        raise ValueError(f"Invalid column_order: {column_order}")

    if order_by is None:
        order_by = left.columns
    # need to cache before ordering to avoid
    # https://github.com/ibis-project/ibis/issues/8442
    left = left.cache().order_by(order_by)
    right = right.cache().order_by(order_by)
    left_records = left.to_pandas().to_dict(orient="records")
    right_records = right.to_pandas().to_dict(orient="records")
    left_records = [make_record_approx(record) for record in left_records]
    right_records = [make_record_approx(record) for record in right_records]
    assert left_records == right_records


def make_record_approx(record: dict) -> dict:
    return {k: make_float_comparable(v) for k, v in record.items()}


def make_float_comparable(val):
    if isinstance(val, dict):
        return {k: make_float_comparable(v) for k, v in val.items()}
    if isinstance(val, (list, tuple)):
        return [make_float_comparable(v) for v in val]
    if pd.isna(val):
        return None
    if isinstance(ibis.literal(val), ir.FloatingValue) and not pd.isna(val):
        return pytest.approx(val, rel=1e-3)
    return val


def get_clusters(
    cluster_id: ibis.Column, *, label: ibis.Column | None = None
) -> set[frozenset]:
    """Convert a label column into a set of clusters.

    Say you have a table of records, and one of the columns
    is an ID that groups records together.
    This function will return a set of frozensets, where each
    frozenset is a cluster of record IDs.

    You can either provide a label column to act as the record IDs,
    or if not given, it will use `ibis.row_number()`.
    """
    if label is None:
        t = cluster_id.name("cluster_id").as_table()
        t = t.mutate(label=ibis.row_number())
        clusters = (
            t.group_by("cluster_id").aggregate(clusters=t.label.collect()).clusters
        )
    else:
        clusters = label.collect().over(group_by=cluster_id)

    def make_hashable(cluster):
        for record_id in cluster:
            if isinstance(record_id, dict):
                yield tuple(record_id.values())
            else:
                yield record_id

    return {frozenset(make_hashable(c)) for c in clusters.execute()}



from __future__ import annotations

import random

import pytest

from mismo.lib.geo import us_census_geocode
from mismo.tests.util import assert_tables_equal

DEFAULT_INS = {
    "id": None,
    "street": None,
    "city": None,
    "state": None,
    "zipcode": None,
}
SCHEMA_INS = {c: "string" for c in DEFAULT_INS}
DEFAULT_EXPECTED = {
    **DEFAULT_INS,
    "census_is_match": False,
    "census_match_type": None,
    "census_street": None,
    "census_city": None,
    "census_state": None,
    "census_zipcode": None,
    "census_latitude": None,
    "census_longitude": None,
}
SCHEMA_EXPECTED = {c: "string" for c in DEFAULT_EXPECTED}
SCHEMA_EXPECTED["census_is_match"] = "boolean"
SCHEMA_EXPECTED["census_latitude"] = "float64"
SCHEMA_EXPECTED["census_longitude"] = "float64"


def make_input(table_factory, records):
    records = [{**DEFAULT_INS, **record} for record in records]
    t = table_factory(records, schema=SCHEMA_INS)
    return t


def make_expected(table_factory, records):
    records = [{**DEFAULT_EXPECTED, **record} for record in records]
    t = table_factory(records, schema=SCHEMA_EXPECTED)
    return t


@pytest.mark.network
def test_us_census_geocode(table_factory):
    GIRDWOOD = {
        "census_is_match": True,
        "census_match_type": "non_exact",
        "census_street": "285 HIGHER TER",
        "census_city": "GIRDWOOD",
        "census_state": "AK",
        "census_zipcode": "99587",
        "census_latitude": 60.954138,
        "census_longitude": -149.115947,
    }
    NORTH_43RD = {
        "census_is_match": True,
        "census_match_type": "exact",
        "census_street": "2114 N 43RD ST",
        "census_city": "SEATTLE",
        "census_state": "WA",
        "census_zipcode": "98103",
        "census_latitude": 47.659183,
        "census_longitude": -122.333537,
    }
    CASKI = {
        "census_is_match": True,
        "census_match_type": "non_exact",
        "census_street": "7258 CASKI CT",
        "census_city": "WASILLA",
        "census_state": "AK",
        "census_zipcode": "99654",
        "census_latitude": 61.497808,
        "census_longitude": -149.650161,
    }
    # not using pytest params because we want to make only one request
    # so the test is faster
    pairs = [
        (
            {"id": "43rd_basic", "street": "2114 North 43rd St", "zipcode": "98103"},
            NORTH_43RD,
        ),
        # can handle duplicates
        (
            {
                "id": "43rd_with_spaces",
                "street": "2114 NORTH 43RD ST  ",
                "zipcode": "98103",
            },
            NORTH_43RD,
        ),
        (
            {"id": "43rd_no_north", "street": "2114 43rd St", "zipcode": "98103"},
            {**NORTH_43RD, "census_match_type": "non_exact"},
        ),
        (
            # APTs are removed!
            {"id": "43rd_apt", "street": "2114 N 43rd St APT 3", "zipcode": "98103"},
            NORTH_43RD,
        ),
        (
            # doesn't exist
            {
                "id": "dne_elm",
                "street": "321   ELM ST",
                "city": "NEW YORK CITY",
                "state": "NEW YORK",
                "zipcode": "10001",
            },
            {"census_is_match": False},
        ),
        (
            {
                "id": "po_box",
                "street": "PO BOX 321",
                "city": "Girdwood",
                "state": "AK",
                "zipcode": "99587",
            },
            {"census_is_match": False},
        ),
        (
            {
                "id": "girdwood_typo_city",
                "street": "285 HIGHER Terrace road",
                "city": "GIRDWOOOD",  # typo
                "state": "AK",
                "zipcode": "99587",
            },
            GIRDWOOD,
        ),
        (
            {
                "id": "girdwood_typo_city_serious",
                "street": "285 HIGHER Terrace road",
                "city": "schmirdwud",  # serious typo
                "state": "ALASKA",
                "zipcode": "99587",
            },
            GIRDWOOD,
        ),
        (
            {
                "id": "girdwood_wrong_city",
                "street": "285 HIGHER Terrace road",
                "city": "seattle",  # totally misleading
                "state": "ALASKA",
                "zipcode": "99587",
            },
            GIRDWOOD,
        ),
        (
            {
                "id": "girdwood_no_state_zip",
                "street": "285 HIGHER Terrace road",
                "city": "GIRDWOoOD",  # typo, no state or zip
            },
            GIRDWOOD,
        ),
        (
            {
                "id": "girdwood_no_state_zip_street_typo",
                "street": "285 HIGHhER Terrace road",  # typo
                "city": "GIRDWOOD",  # no state or zip
            },
            GIRDWOOD,
        ),
        # The lesson in the cases below is that if you aren't getting matches,
        # try REMOVING some info, because that info might be wrong.
        (
            {
                "id": "caski_wrong_zip_with_all",
                "street": "7258 S CASKI CIR",  # Actually is 7258 CASKI CT
                "city": "WASILLA",
                "state": "AK",
                "zipcode": "99623",  # wrong zip
            },
            {"census_is_match": False},
        ),
        (
            {
                "id": "caski_wrong_zip_without_street_type",
                "street": "7258 S CASKI",
                "city": "WASILLA",
                "state": "AK",
                "zipcode": "99623",  # wrong zip
            },
            {"census_is_match": False},
        ),
        (
            {
                "id": "caski_wrong_zip_without_directional",
                "street": "7258 CASKI CIR",  # Without the directional it works!
                "city": "WASILLA",
                "state": "AK",
                "zipcode": "99623",  # wrong zip
            },
            CASKI,
        ),
        (
            {
                "id": "caski",
                "street": "7258 S CASKI CIR",
                "city": "WASILLA",
                "state": "AK",
                "zipcode": "99654",  # with correct zip it works!
            },
            CASKI,
        ),
        (
            {
                "id": "caski_no_zip",
                "street": "7258 S CASKI CIR",
                "city": "WASILLA",
                "state": "AK",
                "zipcode": None,  # without the wrong zip it works!
            },
            CASKI,
        ),
    ]
    ins = [p[0] for p in pairs]
    exp = [{**p[0], **p[1]} for p in pairs]
    inp = make_input(table_factory, ins)
    expected = make_expected(table_factory, exp)
    geocoded = us_census_geocode(inp, chunk_size=2)
    assert_tables_equal(geocoded, expected, column_order="ignore")


@pytest.mark.parametrize(
    "n",
    [
        10,
        100,
        1000,
    ],
)
@pytest.mark.network
def test_benchmark_us_census_geocode(benchmark, table_factory, n: int):
    records = [
        {"street": f"{random.randint(1, 1000)} N 43RD ST", "zipcode": "98103"}
        for _ in range(n)
    ]
    inp = make_input(table_factory, records)
    geocoded = benchmark(us_census_geocode, inp)
    assert geocoded.count().execute() == n



from __future__ import annotations

from mismo.cluster._connected_components import (
    connected_components as connected_components,
)
from mismo.cluster._dashboard import cluster_dashboard as cluster_dashboard
from mismo.cluster._dashboard import clusters_dashboard as clusters_dashboard
from mismo.cluster._eval import (
    adjusted_mutual_info_score as adjusted_mutual_info_score,
)
from mismo.cluster._eval import adjusted_rand_score as adjusted_rand_score
from mismo.cluster._eval import completeness_score as completeness_score
from mismo.cluster._eval import fowlkes_mallows_score as fowlkes_mallows_score
from mismo.cluster._eval import (
    homogeneity_completeness_v_measure as homogeneity_completeness_v_measure,
)
from mismo.cluster._eval import homogeneity_score as homogeneity_score
from mismo.cluster._eval import mutual_info_score as mutual_info_score
from mismo.cluster._eval import (
    normalized_mutual_info_score as normalized_mutual_info_score,
)
from mismo.cluster._eval import rand_score as rand_score
from mismo.cluster._eval import v_measure_score as v_measure_score
from mismo.cluster._metrics import degree as degree
from mismo.cluster._subgraph import degree_dashboard as degree_dashboard



from __future__ import annotations

import asyncio
import csv
import io
import logging
import time
from typing import TYPE_CHECKING, Iterable

import ibis
from ibis.expr import types as ir

from mismo import _util

if TYPE_CHECKING:
    import httpx

# expose this here so we can monkeypatch it in tests
# The API supports max 10k addresses per request
_CHUNK_SIZE = 5_000
_N_CONCURRENT = 16
# If I use a chunk size of 2k,
# with 50 concurrent connections, each request takes 20-30 seconds
# with 100 concurrent connections, each request takes 60-8 seconds
# This implies to me that there are limited resources on the server.
# If you make more requests, the same number of cores are serving them,
# so each request takes longer. But not just linearly,
# probably because of the overhead of context switching.

# results:
# 2k chunk size, 25 concurrent requests: ~15s/req, 5.25 min for 500k rows
# 4k chunk size, 25 concurrent requests: ~30s/req, 5 min for 500k rows
# 4k chunk size, 15 concurrent requests: ~22s/req, 3.75 min for 500k rows
# 5k chunk size, 15 concurrent requests: ~??s/req, 3.25 min for 500k rows
# 5k chunk size, 15 concurrent requests: ~26s/req, 3.5 min for 500k rows
# 10k chunk size, 15 concurrent requests: ~65s/req, 4 min for 500k rows
# 5k chunk size, 10 concurrent requests: ~25s/req, 4.5 min for 500k rows
# 5k chunk size, 20 concurrent requests: ~35s/req, 4.5 min for 500k rows
# 5k chunk size, 16 concurrent requests: ~30s/req, 3.5 min for 500k rows


logger = logging.getLogger(__name__)


def us_census_geocode(
    t: ir.Table,
    format: str = "census_{name}",
    *,
    benchmark: str | None = None,
    vintage: str | None = None,
    chunk_size: int | None = None,
    n_concurrent: int | None = None,
) -> ir.Table:
    """Geocode US physical addresses using the US Census Bureau's geocoding service.

    Uses the batch geocoding API from https://geocoding.geo.census.gov/geocoder.
    This only works for US physical addresses. PO Boxes are not supported.
    "APT 123", "UNIT B", etc are not included in the results, so you will need
    to extract those before geocoding.

    Before geocoding, this function normalizes the input addresses and deduplicates
    them, so if your input table has 1M rows, but only 100k unique addresses,
    it will only send those 100k addresses to the API.

    This took about 7 minutes to geocode 1M unique addresses in my tests.

    Parameters
    ----------
    t:
        A table of addresses to geocode. Must have the schema:
        - street: string, the street address.
        - city: string, the city name.
        - state: string, the state name.
        - zipcode: string, the ZIP code.
    format:
        The format to use for the output column names. See the Returns section.
    benchmark:
        The geocoding benchmark to use. Default is "Public_AR_Current".
    vintage:
        The geocoding vintage to use. Default is "Current_Current".
    chunk_size:
        The number of addresses to geocode in each request. Default is 5000.
        The maximum allowed by the API is 10_000.
        This number was tuned experimentally, you probably don't need to change ir.
    n_concurrent:
        The number of concurrent requests to make. Default is 16.
        This number was tuned experimentally, you probably don't need to change ir.

    Returns
    -------
    geocoded:
        The input table, with the following additional columns:
        - is_match: bool, whether the address was successfully matched.
          If False, all the other columns will be NULL.
        - match_type: string, the type of match. eg "exact", "non_exact"
        - street: string, the normalized street address
        - city: string, the normalized city name
        - state: string, the normalized 2 letter state code
        - zipcode: string, the 5 digit ZIP code
        - latitude: float64, the latitude of the matched address
        - longitude: float64, the longitude of the matched address
        Each of these columns is named according to the `format` parameter.
        For example, if `format` is "census_{name}", the columns will be named
        "census_is_match", "census_match_type", "census_street", etc.
        The order of the results is not guaranteed to match the input order.

    """
    if chunk_size is None:
        chunk_size = _CHUNK_SIZE
    if chunk_size > 10_000:
        raise ValueError("chunk_size must be <= 10_000")
    if n_concurrent is None:
        n_concurrent = _N_CONCURRENT
    t = t.mutate(__row_number=ibis.row_number())
    normed = _normed_addresses(t)
    deduped, restore = _dedupe(normed)
    gc = _geocode(
        deduped,
        benchmark=benchmark,
        vintage=vintage,
        chunk_size=chunk_size,
        n_concurrent=n_concurrent,
    )
    unduped = restore(gc)
    unduped = unduped.rename(
        {
            format.format(name=col): col
            for col in unduped.columns
            if col != "__row_number"
        }
    )
    re_joined = t.inner_join(unduped, "__row_number").drop("__row_number")
    return re_joined


def _dedupe(t: ir.Table) -> tuple[ir.Table, callable]:
    keys = ["street", "city", "state", "zipcode"]
    api_id = ibis.dense_rank().over(ibis.window(order_by=keys))
    # same as pandas.DataFrame.groupby(keys).ngroup()
    with_group_id = t.mutate(api_id=api_id)
    deduped = with_group_id.select(
        "api_id", "street", "city", "state", "zipcode"
    ).distinct()
    # need to cache this, otherwise if you look at deduped[:100], deduped[100:200],
    # etc, it will recompute deduped each time, and since the order is not guaranteed,
    # you might get api_id 1 both times!
    deduped = deduped.cache()
    restore_map = with_group_id.select("__row_number", "api_id")

    def restore(ded: ir.Table) -> ir.Table:
        return restore_map.left_join(ded, "api_id").drop("api_id", "api_id_right")

    return deduped, restore


def _geocode(
    t: ir.Table,
    *,
    benchmark,
    vintage,
    chunk_size: int,
    n_concurrent: int,
) -> ir.Table:
    t = t.cache()
    sub_tables = chunk_table(t, max_size=chunk_size)
    byte_chunks = (_table_to_csv_bytes(sub) for sub in sub_tables)
    client = _make_client()
    sem = asyncio.Semaphore(n_concurrent)
    requests = [
        dict(bytes=b, benchmark=benchmark, vintage=vintage) for b in byte_chunks
    ]
    logger.debug(
        f"Geocoding {t.count().execute()} addresses in {len(requests)} chunks of size {chunk_size}"  # noqa
    )
    responses = _make_requests(requests, client=client, sem=sem)
    tables = (_text_to_table(resp_text) for resp_text in responses)
    result = ibis.union(*tables)
    result = _post_process_table(result)
    return result


def _normed_addresses(t: ir.Table) -> ir.Table:
    t = t.select(
        "__row_number",
        street=t.street.strip().upper(),
        city=t.city.strip().upper(),
        state=t.state.strip().upper(),
        zipcode=t.zipcode.strip().upper(),
    )
    return t


def chunk_table(t: ir.Table, max_size: int) -> Iterable[ir.Table]:
    n = t.count().execute()
    i = 0
    while i < n:
        yield t[i : i + max_size]
        i += max_size


def _table_to_csv_bytes(t: ir.Table) -> bytes:
    t = t.select("api_id", "street", "city", "state", "zipcode")
    df = t.to_pandas()
    return df.to_csv(index=False, header=False).encode("utf-8")


def _make_client() -> httpx.AsyncClient:
    with _util.optional_import("httpx"):
        import httpx
    timeout = httpx.Timeout(10, read=150, pool=10000)
    return httpx.AsyncClient(timeout=timeout)


def _make_requests(
    requests: Iterable[dict],
    *,
    client: httpx.AsyncClient | None = None,
    sem: asyncio.Semaphore | None = None,
):
    if client is None:
        client = _make_client()
    if sem is None:
        sem = asyncio.Semaphore(_N_CONCURRENT)

    async def _async_make_requests() -> list[str]:
        async with client:
            responses = [
                _make_request(client, sem, chunk_id=i, **req)
                for i, req in enumerate(requests)
            ]
            return await asyncio.gather(*responses)

    return _asyncio_run_with_nest_asyncio(_async_make_requests())


async def _make_request(
    client: httpx.AsyncClient,
    sem: asyncio.Semaphore,
    *,
    chunk_id: int,
    bytes: bytes,
    benchmark: str | None = None,
    vintage: str | None = None,
) -> str:
    URL = "https://geocoding.geo.census.gov/geocoder/locations/addressbatch"
    data = {
        "benchmark": benchmark or "Public_AR_Current",
        "vintage": vintage or "Current_Current",
    }
    files = {"addressFile": ("addresses.csv", bytes)}

    # Sometimes I get a 502 error:
    # While attempting to geocode your batch input, the Census Geocoder
    # working as a gateway to get a response needed to handle the request,
    # got an invalid response. This may be caused by the fact that the
    # Census Geocoder is unable to connect to a service it is dependent
    # response received was invalid. Please verify your batch file and
    # resubmit your batch query again later.

    # Then if I resubmit the same request, it works. So I don't think the issue
    # is the format of the request. I think the Census Geocoder is just flaky.
    # Nothing to do but retry.

    async def f():
        async with sem:
            logger.debug(f"Geocoding chunk {chunk_id}")
            start = time.monotonic()
            resp = await client.post(URL, data=data, files=files)
            if resp.status_code != 200:
                raise RuntimeError(f"Failed to geocode chunk {chunk_id}: {resp.text}")
            end = time.monotonic()
            logger.debug(f"Geocoded chunk {chunk_id} in {end - start:.2f} seconds")
            return resp.text

    return await _with_retries(f, chunk_id=chunk_id, max_retries=3)


async def _with_retries(f, *, chunk_id, max_retries: int):
    for i in range(max_retries):
        try:
            return await f()
        except Exception as e:
            logger.warning(
                f"Retrying chunk {chunk_id} {i + 1}/{max_retries} after error: {e!r}"
            )
            if i == max_retries - 1:
                raise RuntimeError(f"Failed to geocode chunk {chunk_id}: {e!r}") from e
            else:
                # exponential backoff
                # 0.5, 1, 2, 4, 8
                await asyncio.sleep(0.5 * 2**i)


def _text_to_table(text: str) -> ir.Table:
    _RAW_SCHEMA = ibis.schema(
        {
            "api_id": "string",
            "address": "string",
            "match": "string",
            "matchtype": "string",
            "parsed": "string",
            "coordinate": "string",
            "tigerlineid": "string",
            "side": "string",
        }
    )
    with io.StringIO(text) as f:
        records = list(csv.DictReader(f, fieldnames=list(_RAW_SCHEMA.keys())))
    if not records:
        records = {col: [] for col in _RAW_SCHEMA}
    return ibis.memtable(records, schema=_RAW_SCHEMA)


def _post_process_table(t: ir.Table) -> ir.Table:
    lonlat = t.coordinate.split(",")
    lon, lat = lonlat[0].cast("float64"), lonlat[1].cast("float64")
    # pattern is r"(.*), (.*), (.*), (.*)" but splitting on ", " is (probably?) faster
    parts: ir.ArrayColumn = t.parsed.split(", ")
    street = parts[0].strip()
    city = parts[1].strip()
    state = parts[2].strip()
    zipcode = parts[3].strip()
    return t.select(
        api_id=t.api_id.cast("int64"),
        is_match=t.match == "Match",
        match_type=t.matchtype.lower(),
        street=street,
        city=city,
        state=state,
        zipcode=zipcode,
        latitude=lat,
        longitude=lon,
    )


def _asyncio_run_with_nest_asyncio(coro):
    """asyncio.run(), but can handle nested loops as in Jupyter."""
    try:
        return asyncio.run(coro)
    except RuntimeError as e:
        if "asyncio.run() cannot be called from a running event loop" not in str(e):
            raise
        import nest_asyncio

        nest_asyncio.apply()
        return asyncio.run(coro)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

import ibis

from mismo import _common, _typing
from mismo.types._linked_table import LinkedTable
from mismo.types._links_table import LinksTable

if TYPE_CHECKING:
    import altair as alt
    from ibis.expr import types as ir


class Linkage:
    """A dataclass of two Tables of records ([LinkedTables][mismo.LinkedTable]) and a Table of links ([LinksTable][mismo.LinksTable]) between them.

    Each record in `left` can be linked from 0 to N records in `right`, and vice versa.

    See Also
    --------
    The [Diff][mismo.Diff] dataclass, for representing the special case
    where each record is linked to at most one other record.
    """  # noqa: E501

    def __init__(
        self, *, left: ibis.Table, right: ibis.Table, links: ibis.Table
    ) -> None:
        """Create from two Tables and a Table of links between them.

        Parameters
        ----------
        left
            A Table of records, with at least a column 'record_id'.
        right
            A Table of records, with at least a     column 'record_id'.
        links
            A Table of links between the two tables.
            Must have columns 'record_id_l' and 'record_id_r', which refer to the
            'record_id' columns in `left` and `right`, respectively.
            May have other columns.
            May not have duplicate (record_id_l, record_id_r) pairs.
        """
        _common.check_tables_and_links(left, right, links)

        self._left, self._right = LinkedTable.make_pair(
            left=left, right=right, links=links
        )
        self._raw_links = links

    @property
    def left(self) -> LinkedTable:
        """The left Table."""
        return self._left

    @property
    def right(self) -> LinkedTable:
        """The right Table."""
        return self._right

    @property
    def links(self) -> LinksTable:
        """
        A table of (record_id_l, record_id_r, <other attributes>...) that link `left` and `right`.
        """  # noqa: E501
        return LinksTable(self._raw_links, left=self._left, right=self._right)

    @classmethod
    def from_join_condition(
        cls,
        left: ibis.Table,
        right: ibis.Table,
        condition: Any,
    ) -> _typing.Self:
        """
        Create a Linkage from two Tables and a join condition.

        Parameters
        ----------
        left
            A Table of records, with at least a column 'record_id'.
        right
            A Table of records, with at least a column 'record_id'.
        condition
            A join condition, such as a boolean expression or an ibis expression.
            See [mismo.join_condition][] for more details.

        Returns
        -------
            A Linkage object.
        """
        links = LinksTable.from_join_condition(left, right, condition)
        return cls(left=left, right=right, links=links)

    def cache(self) -> _typing.Self:
        """Cache left, right, and links for faster subsequent access."""
        return self.__class__(
            left=self.left.cache(), right=self.right.cache(), links=self.links.cache()
        )

    def to_parquets(self, directory: str | Path, /, *, overwrite: bool = False) -> None:
        """
        Write left, right, and links to parquet files in the given directory.
        """
        d = Path(directory)
        d.mkdir(parents=True, exist_ok=True)

        def write(t: ir.Table, name: str):
            p = d / f"{name}.parquet"
            if p.exists() and not overwrite:
                raise FileExistsError(f"{p} already exists")
            t.to_parquet(p)

        write(self.left, "left")
        write(self.right, "right")
        write(self.links, "links")

    @classmethod
    def from_parquets(
        cls, directory: str | Path, /, *, backend: ibis.BaseBackend | None = None
    ) -> _typing.Self:
        """Create a Linkage by reading parquets from the given directory."""
        if backend is None:
            backend = ibis
        d = Path(directory)
        return cls(
            left=backend.read_parquet(d / "left.parquet"),
            right=backend.read_parquet(d / "right.parquet"),
            links=backend.read_parquet(d / "links.parquet"),
        )

    def __repr__(self):
        return f"{self.__class__.__name__}<left={self.left.count().execute():_}, right={self.right.count().execute():_}, links={self.links.count().execute():_}>"  # noqa: E501

    def copy(
        self,
        *,
        left: LinkedTable | None = None,
        right: LinkedTable | None = None,
        links: LinksTable | None = None,
    ) -> Linkage:
        """
        Create a new Linkage, optionally replacing the left, right, and links tables.
        """
        return self.__class__(
            left=left if left is not None else self.left,
            right=right if right is not None else self.right,
            links=links if links is not None else self.links,
        )

    def link_counts_chart(self) -> alt.Chart:
        """
        A side by side altair Chart of `left.link_counts().chart()` and `right.link_counts().chart()`.

        ```plaintext
        Number of           Left Table               Number of    Right Table
          Records                                      Records
                |    █                                       |    █
        100,000 | █  █                                       |    █
                | █  █                                10,000 |    █
                | █  █  █                                    |    █
         10,000 | █  █  █                                    |    █  █
                | █  █  █                                    | █  █  █
                | █  █  █                              1,000 | █  █  █
          1,000 | █  █  █  █                                 | █  █  █
                | █  █  █  █  █  █                           | █  █  █
                | █  █  █  █  █  █  █                        | █  █  █  █
            100 | █  █  █  █  █  █  █  █  █              100 | █  █  █  █
                | 0  1  2  3  4 10 12 14 23                  | 0  1  2  3
                Number of Links                              Number of Links
        ```
        """  # noqa: E501
        import altair as alt

        left = self.left.link_counts().chart()
        right = self.right.link_counts().chart().properties(title="Right Table")
        subtitle = left.title.subtitle
        left = left.properties(title=alt.TitleParams("Left Table", anchor="middle"))
        right = right.properties(title=alt.TitleParams("Right Table", anchor="middle"))
        return alt.hconcat(left, right).properties(
            title=alt.TitleParams(
                "Number of Records by Link Count", subtitle=subtitle, anchor="middle"
            )
        )


Linkish = TypeVar("T", bound=LinksTable | Linkage)


# TODO: IDK if this deserves to be its own function,
# or if this should just get covered in a HowTo guide,
# and users should implement it themselves.
def filter_links(links_or_linkage: Linkish, condition: ir.BooleanValue) -> Linkish:
    """
    Create a new Linkage/LinksTable, filtered by the given condition.

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> links = ibis.memtable(
    ...     [
    ...         (1, "a", 0.45),
    ...         (1, "b", 0.67),
    ...         (2, "c", 0.23),
    ...         (2, "c", 0.87),
    ...         (3, "d", 0.12),
    ...         (4, "d", 0.97),
    ...     ],
    ...     schema={
    ...         "record_id_l": "int64",
    ...         "record_id_r": "string",
    ...         "score": "float64",
    ...     },
    ... )

    We only want to keep links that are above a certain score.

    >>> filter_links(links, ibis._.score > 0.5)
    ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━┓
    ┃ record_id_l ┃ record_id_r ┃ score   ┃
    ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━┩
    │ int64       │ string      │ float64 │
    ├─────────────┼─────────────┼─────────┤
    │           1 │ b           │    0.67 │
    │           2 │ c           │    0.87 │
    │           4 │ d           │    0.97 │
    └─────────────┴─────────────┴─────────┘

    Or, say we are doing a lookup into a clean table (left) from a
    new set of dirty data (right).
    We want to only include links that are unambiguous,
    eg where each record in right is linked to at most one record in left.

    >>> filter_links(
    ...     links, (ibis._.record_id_l.nunique() == 1).over(group_by="record_id_r")
    ... ).order_by("score")
    ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━┓
    ┃ record_id_l ┃ record_id_r ┃ score   ┃
    ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━┩
    │ int64       │ string      │ float64 │
    ├─────────────┼─────────────┼─────────┤
    │           2 │ c           │    0.23 │
    │           1 │ a           │    0.45 │
    │           1 │ b           │    0.67 │
    │           2 │ c           │    0.87 │
    └─────────────┴─────────────┴─────────┘

    Or, see how there are two links between 2 and c.
    We only want to keep the one with the highest score, per each record in right.

    >>> filter_links(
    ...     links, (ibis._.score == ibis._.score.max()).over(group_by="record_id_r")
    ... ).order_by("score")
    ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━┓
    ┃ record_id_l ┃ record_id_r ┃ score   ┃
    ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━┩
    │ int64       │ string      │ float64 │
    ├─────────────┼─────────────┼─────────┤
    │           1 │ a           │    0.45 │
    │           1 │ b           │    0.67 │
    │           2 │ c           │    0.87 │
    │           4 │ d           │    0.97 │
    └─────────────┴─────────────┴─────────┘

    Or, perhaps say we have many different linking methods,
    eg link where the names match, or where the addresses match.
    We only want to keep links where at least two of the linking methods agree.

    >>> filter_links(
    ...     links,
    ...     (ibis._.count() >= 2).over(group_by=("record_id_l", "record_id_r")),
    ... )
    ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━┓
    ┃ record_id_l ┃ record_id_r ┃ score   ┃
    ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━┩
    │ int64       │ string      │ float64 │
    ├─────────────┼─────────────┼─────────┤
    │           2 │ c           │    0.23 │
    │           2 │ c           │    0.87 │
    └─────────────┴─────────────┴─────────┘
    """  # noqa: E501
    if isinstance(links_or_linkage, Linkage):
        return links_or_linkage.copy(
            links=filter_links(links_or_linkage.links, condition)
        )
    else:
        return links_or_linkage.filter(condition)



from __future__ import annotations

from typing import Iterable, Mapping, overload

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo._datasets import Datasets


@overload
def degree(
    *,
    links: ir.Table,
    records: ir.Table | None,
) -> ir.Table: ...


@overload
def degree(
    *,
    links: ir.Table,
    records: Iterable[ir.Table] | Mapping[str, ir.Table],
) -> Datasets: ...


def degree(
    *,
    links: ir.Table,
    records: ir.Table | Iterable[ir.Table] | Mapping[str, ir.Table] | None = None,
) -> ir.Table | Datasets:
    """Label records with their degree (number of links to other records).

    This is the graph theory definition of degree, i.e. the number of vertices
    coming into or out of a vertex. In this case, the number of links coming
    into or out of a record.

    Parameters
    ----------
    links :
        A table of edges with at least columns (record_id_l, record_id_r).
    records :
        Table(s) of records with at least the column `record_id`, or None.

    Returns
    -------
    result
        If `records` is None, a Table will be returned with columns
        `record_id` and `degree:uint64` that maps record_id to a degree.
        If `records` is a single Table, that table will be returned
        with a `degree:uint64` column added.
        If an iterable/mapping of Tables is given, a `Datasets` will be returned,
        with a `component` column added to each contained Table.
    """
    l1 = links.select(record_id="record_id_l", other="record_id_r")
    l2 = links.select(record_id="record_id_r", other="record_id_l")
    u = ibis.union(l1, l2, distinct=True)
    lookup = u.group_by("record_id").agg(degree=_.count())
    if records is None:
        return lookup
    ds = Datasets(records).map(
        lambda _name, t: t.left_join(lookup, "record_id")
        .mutate(degree=_.degree.fill_null(0))
        .drop("record_id_right")
    )
    if len(ds) == 1:
        return ds[0]
    return ds



"""Houses domain-specific stuff, such as for names, addresses, locations, etc."""

from __future__ import annotations

from mismo.lib import email as email
from mismo.lib import geo as geo
from mismo.lib import name as name
from mismo.lib import phone as phone



from __future__ import annotations

import dataclasses
import math
from typing import Callable

import ibis
from ibis import Deferred
from ibis.expr import types as ir

from mismo import _util, linkage, types


def distance_km(
    *,
    lat1: ir.FloatingValue,
    lon1: ir.FloatingValue,
    lat2: ir.FloatingValue,
    lon2: ir.FloatingValue,
) -> ir.FloatingValue:
    """The distance between two points on the Earth's surface, in kilometers.

    Parameters
    ----------
    lat1:
        The latitude of the first point.
    lon1:
        The longitude of the first point.
    lat2:
        The latitude of the second point.
    lon2:
        The longitude of the second point.

    Returns
    -------
    distance:
        The distance between the two points, in kilometers.
    """

    def radians(degrees):
        scaling = math.pi / 180
        return degrees * scaling

    lat1 = radians(lat1)
    lon1 = radians(lon1)
    lat2 = radians(lat2)
    lon2 = radians(lon2)

    def haversine(theta: ir.FloatingValue) -> ir.FloatingValue:
        return (theta / 2).sin() ** 2

    a = haversine(lat2 - lat1) + lat1.cos() * lat2.cos() * haversine(lon2 - lon1)
    R_earth = 6371.0  # km
    # parens for optimization, to make sure the division happens on the python side
    return (R_earth * 2) * a.sqrt().asin()


@dataclasses.dataclass(frozen=True)
class CoordinateBlocker:
    """Blocks two locations together if they are within a certain distance.

    This isn't precise, and can include pairs that are actually up to about 2x
    larger than the given threshold.
    This is because we use a simple grid to bin the coordinates, so
    1. This isn't accurate near the poles, and
    2. This isn't accurate near the international date line (longitude 180/-180).
    3. If two coords fall within opposite corners of the same grid cell, they
        will be blocked together even if they are further apart than the
        precision, due to the diagonal distance being longer than the horizontal
        or vertical distance.

    Examples
    --------
    >>> import ibis
    >>> from mismo.lib.geo import CoordinateBlocker
    >>> ibis.options.interactive = True
    >>> conn = ibis.duckdb.connect()
    >>> left = conn.create_table(
    ...     "left",
    ...     [
    ...         {
    ...             "record_id": 0,
    ...             "latlon": {"lat": 61.1547800, "lon": -150.0677490},
    ...         }
    ...     ],
    ... )
    >>> right = conn.create_table(
    ...     "right",
    ...     [
    ...         {
    ...             "record_id": 4,
    ...             "latitude": 61.1582056,
    ...             "longitude": -150.0584552,
    ...         },
    ...         {
    ...             "record_id": 5,
    ...             "latitude": 61.1582056,
    ...             "longitude": 0,
    ...         },
    ...         {
    ...             "record_id": 6,
    ...             "latitude": 61.1547800,
    ...             "longitude": -150,
    ...         },
    ...     ],
    ... )
    >>> blocker = CoordinateBlocker(
    ...     distance_km=1,
    ...     name="within_1_km",
    ...     left_coord="latlon",
    ...     right_lat="latitude",
    ...     right_lon="longitude",
    ... )
    >>> blocker(left, right).links
    ┏━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━━┓
    ┃ record_id_l ┃ latlon_l                      ┃ record_id_r ┃ latitude_r ┃ longitude_r ┃
    ┡━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━━┩
    │ int64       │ struct<lat: float64, lon:     │ int64       │ float64    │ float64     │
    │             │ float64>                      │             │            │             │
    ├─────────────┼───────────────────────────────┼─────────────┼────────────┼─────────────┤
    │             │ {                             │             │            │             │
    │           0 │     'lat': 61.15478,          │           4 │  61.158206 │ -150.058455 │
    │             │     'lon': -150.067749        │             │            │             │
    │             │ }                             │             │            │             │
    └─────────────┴───────────────────────────────┴─────────────┴────────────┴─────────────┘
    """  # noqa: E501

    distance_km: float | int
    """
    The (approx) max distance in kilometers that two coords will be blocked together.
    """
    name: str | None = None
    """The name of the blocker."""
    coord: str | Deferred | Callable[[ir.Table], ir.StructColumn] | None = None
    """The column in both tables containing the `struct<lat: float, lon: float>` coordinates."""  # noqa: E501
    lat: str | Deferred | Callable[[ir.Table], ir.FloatingColumn] | None = None
    """The column in both tables containing the latitude coordinates."""
    lon: str | Deferred | Callable[[ir.Table], ir.FloatingColumn] | None = None
    """The column in both tables containing the longitude coordinates."""
    left_coord: str | Deferred | Callable[[ir.Table], ir.StructColumn] | None = None
    """The column in the left tables containing the `struct<lat: float, lon: float>` coordinates."""  # noqa: E501
    right_coord: str | Deferred | Callable[[ir.Table], ir.StructColumn] | None = None
    """The column in the right tables containing the `struct<lat: float, lon: float>` coordinates."""  # noqa: E501
    left_lat: str | Deferred | Callable[[ir.Table], ir.FloatingColumn] | None = None
    """The column in the left tables containing the latitude coordinates."""
    left_lon: str | Deferred | Callable[[ir.Table], ir.FloatingColumn] | None = None
    """The column in the left tables containing the longitude coordinates."""
    right_lat: str | Deferred | Callable[[ir.Table], ir.FloatingColumn] | None = None
    """The column in the right tables containing the latitude coordinates."""
    right_lon: str | Deferred | Callable[[ir.Table], ir.FloatingColumn] | None = None
    """The column in the right tables containing the longitude coordinates."""

    def __post_init__(self):
        ok_subsets = [
            {"coord"},
            {"left_coord", "right_coord"},
            {"left_coord", "right_lat", "right_lon"},
            {"left_lat", "left_lon", "right_coord"},
            {"left_lat", "left_lon", "right_lat", "right_lon"},
            {"lat", "lon"},
            {"lat", "right_lat", "right_lon"},
            {"left_lat", "left_lon", "lon"},
        ]
        options = [
            "coord",
            "left_coord",
            "right_coord",
            "lat",
            "lon",
            "left_lat",
            "left_lon",
            "right_lat",
            "right_lon",
        ]
        present = {k for k in options if getattr(self, k) is not None}
        if present not in ok_subsets:
            ok_subsets_str = "\n".join("- " + str(s) for s in ok_subsets)
            raise ValueError(
                "You must specify exactly one of the following subsets of options:\n"
                + ok_subsets_str
                + f"\nYou provided:\n{present}"
            )

    def _get_cols(
        self, left: ir.Table, right: ir.Table
    ) -> tuple[
        ir.FloatingColumn, ir.FloatingColumn, ir.FloatingColumn, ir.FloatingColumn
    ]:
        if self.coord is not None:
            left
            left_coord = _util.get_column(left, self.coord, on_many="struct")
            right_coord = _util.get_column(right, self.coord, on_many="struct")
            return (left_coord.lat, left_coord.lon, right_coord.lat, right_coord.lon)
        if self.left_coord is not None:
            left_coord = _util.get_column(left, self.left_coord, on_many="struct")
            left_lat = left_coord.lat
            left_lon = left_coord.lon
        if self.right_coord is not None:
            right_coord = _util.get_column(right, self.right_coord, on_many="struct")
            right_lat = right_coord.lat
            right_lon = right_coord.lon
        if self.lat is not None:
            left_lat = _util.get_column(left, self.lat)
            right_lat = _util.get_column(right, self.lat)
        if self.lon is not None:
            left_lon = _util.get_column(left, self.lon)
            right_lon = _util.get_column(right, self.lon)
        if self.left_lat is not None:
            left_lat = _util.get_column(left, self.left_lat)
        if self.left_lon is not None:
            left_lon = _util.get_column(left, self.left_lon)
        if self.right_lat is not None:
            right_lat = _util.get_column(right, self.right_lat)
        if self.right_lon is not None:
            right_lon = _util.get_column(right, self.right_lon)
        return left_lat, left_lon, right_lat, right_lon

    def __join_condition__(self, left: ir.Table, right: ir.Table) -> ir.BooleanValue:
        left_lat, left_lon, right_lat, right_lon = self._get_cols(left, right)
        # We have to use a grid size of ~3x the precision to avoid
        # two points falling right on either side of a grid cell boundary
        grid_size = self.distance_km * 3
        left_hashed = _bin_lat_lon(left_lat, left_lon, grid_size)
        right_hashed = _bin_lat_lon(right_lat, right_lon, grid_size)
        return left_hashed == right_hashed

    def __call__(self, left: ir.Table, right: ir.Table) -> linkage.Linkage:
        links = types.LinksTable.from_join_condition(left, right, self)
        return linkage.Linkage(left=left, right=right, links=links)


def _bin_lat_lon(
    lat: ir.FloatingValue, lon: ir.FloatingValue, grid_size_km: float | int
) -> ir.StructValue:
    """Bin a latitude or longitude to a grid of a given precision.

    Say you have two coordinates, (lat1, lon1) and (lat2, lon2), and you
    want to see if they are within, say 15km of each other. You can bin
    them both to a grid of 15km precision and compare the resulting
    hashed values. If they are the same, the coordinates are within 15km
    of each other.
    """
    km_per_lat, km_per_lon = _km_per_degree(lat)
    step_size_lat = grid_size_km / km_per_lat
    step_size_lon = grid_size_km / km_per_lon

    result = ibis.struct(
        {
            # once https://github.com/ibis-project/ibis/pull/10353 is fixed
            # we can use lat // step_size_lat
            "lat_hash": (lat / step_size_lat).floor().cast(int),
            "lon_hash": (lon / step_size_lon).floor().cast(int),
        }
    )
    both_null = lat.isnull() & lon.isnull()
    return both_null.ifelse(ibis.null(), result)


def _km_per_degree(lat: ir.FloatingValue) -> tuple[float, ir.FloatingValue]:
    # Radius of the Earth in kilometers
    R = 6371.0
    lat_rad = lat * (math.pi / 180)
    # This is a constant value at any given latitude.
    km_per_lat = (math.pi * R) / 180
    # This varies based on the latitude.
    # Parens for optimization, to make sure the division happens on the python side.
    km_per_lon = lat_rad.cos() * ((math.pi * R) / 180)
    return km_per_lat, km_per_lon



from __future__ import annotations

from collections.abc import Iterable

import ibis

from mismo.types._table_wrapper import TableWrapper


class UnionTable(TableWrapper):
    """
    A Table whose rows are the non-unique union of the rows from all sub-Tables.
    """

    def __init__(self, tables: Iterable[ibis.Table]) -> None:
        """
        Create a UnionTable from an iterable of ibis Tables.
        """
        tables = tuple(tables)
        if not tables:
            raise ValueError("At least one table must be provided.")
        unioned_table = ibis.union(*tables, distinct=False)
        super().__init__(unioned_table)
        object.__setattr__(self, "_tables", tables)

    @property
    def tables(self) -> tuple[ibis.Table, ...]:
        """
        The tuple of underlying ibis Tables.
        """
        return self._tables

    # skip type hints to inherit from the grandparent ibis.Table
    def filter(self, *predicates):
        return UnionTable(t.filter(*predicates) for t in self.tables)

    # skip type hints to inherit from the grandparent ibis.Table
    def select(self, *exprs, **named_exprs):
        return UnionTable(t.select(*exprs, **named_exprs) for t in self.tables)

    # skip type hints to inherit from the grandparent ibis.Table
    def mutate(self, *args, **kwargs):
        return UnionTable(t.mutate(*args, **kwargs) for t in self.tables)

    # skip type hints to inherit from the grandparent ibis.Table
    def rename(self, method=None, /, **substitutions):
        return UnionTable(t.rename(method, **substitutions) for t in self.tables)

    # skip type hints to inherit from the grandparent ibis.Table
    def drop(self, *fields):
        return UnionTable(t.drop(*fields) for t in self.tables)



from __future__ import annotations

from mismo.linker._basic import EmptyLinker as EmptyLinker
from mismo.linker._basic import FullLinker as FullLinker
from mismo.linker._basic import UnnestLinker as UnnestLinker
from mismo.linker._basic import empty_linkage as empty_linkage
from mismo.linker._basic import full_linkage as full_linkage
from mismo.linker._common import Linker as Linker
from mismo.linker._id_linker import IDLinker as IDLinker
from mismo.linker._join_linker import JoinLinker as JoinLinker
from mismo.linker._key_linker import KeyLinker as KeyLinker
from mismo.linker._or_linker import OrLinker as OrLinker



from __future__ import annotations

import ibis
import pytest

import mismo


def _n_pairs(n: int, k: int) -> int:
    return (k - 1) * (n - k) + sum(range(k))


def _id_func(nk):
    n, k = nk
    return f"n={n},k={k},npairs={_n_pairs(n, k):,}"


@pytest.mark.parametrize(
    "nk",
    [
        (100_000, 5),
        (1_000_000, 5),
    ],
    ids=_id_func,
)
def test_benchmark_block_array(backend, benchmark, nk):
    n, k = nk
    t = ibis.memtable(
        {
            "record_id": list(range(n)),
            "vals": [list(range(i, i + k)) for i in range(n)],
        }
    )
    t = backend.create_table("t", t)
    i = 0

    def f():
        b = mismo.linker.UnnestLinker("vals")(t, t)
        # do this to prevent caching
        nonlocal i
        b = backend.create_table(f"b{i}", b)
        i += 1
        return b

    result = benchmark(f)
    result = result.order_by(["record_id_l", "record_id_r"])
    first_few_expected = [
        (0, 1),
        (0, 2),
        (0, 3),
        (0, 4),
        (1, 2),
        (1, 3),
        (1, 4),
        (1, 5),
        (2, 3),
        (2, 4),
    ]
    first_few = list(
        result[["record_id_l", "record_id_r"]]
        .head(len(first_few_expected))
        .to_pandas()
        .itertuples(index=False, name=None)
    )
    assert first_few_expected == list(first_few)
    n_expected = _n_pairs(n, k)
    assert n_expected == result.count().execute()



from __future__ import annotations

import logging
from typing import Iterable

from ibis import _
from ibis.expr import types as ir

from mismo.compare import LevelComparer

from . import _train
from ._weights import ComparerWeights, Weights

logger = logging.getLogger(__name__)


def train_using_em(
    comparers: Iterable[LevelComparer],
    left: ir.Table,
    right: ir.Table,
    *,
    max_pairs: int | None = None,
) -> Weights:
    """Train weights on unlabeled data using an expectation maximization algorithm.

    Parameters
    ----------
    comparers
        The comparers to train.
    left
        The left dataset.
    right
        The right dataset.
    max_pairs
        The maximum number of pairs to sample.
        If None, all pairs are used.

    Returns
    -------
    Weights
        The estimated weights for each comparer.
    """
    links = _train.sample_all_links(left, right, max_pairs=max_pairs)
    links = links.with_both()
    for c in comparers:
        links = c(links)
    links = links.select([c.name for c in comparers])
    links = links.cache()
    weights = _initial_weights(comparers, links)
    for i in range(5):
        logger.info("EM iteration {i}, starting weights: {weights}", i, weights)
        scored = weights.score_compared(links)
        is_match = _.odds >= 10
        matches = scored.filter(is_match)
        nonmatches = scored.filter(~is_match)
        weights = _weights_from_matches_nonmatches(comparers, matches, nonmatches)
    return weights


def _initial_weights(comparers: Iterable[LevelComparer], labels: ir.Table) -> Weights:
    return Weights(_initial_comparer_weights(c, labels[c.name]) for c in comparers)


def _initial_comparer_weights(
    comparer: LevelComparer, labels: ir.IntegerColumn
) -> ComparerWeights:
    n_levels = len(comparer.levels)
    ms = [1 / n_levels] * n_levels
    us = _train.level_proportions(comparer.levels, labels)
    return _train.make_weights(comparer, ms, us)


def _weights_from_matches_nonmatches(
    comparers: Iterable[LevelComparer], matches: ir.Table, nonmatches: ir.Table
) -> Weights:
    return Weights(
        [
            _comparer_weights_from_matches_nonmatches(
                comp, matches[comp.name], nonmatches[comp.name]
            )
            for comp in comparers
        ]
    )


def _comparer_weights_from_matches_nonmatches(
    comparer: LevelComparer,
    match_labels: ir.IntegerColumn | ir.StringColumn,
    nonmatch_labels: ir.IntegerColumn | ir.StringColumn,
) -> ComparerWeights:
    ms = _train.level_proportions(comparer.levels, match_labels)
    us = _train.level_proportions(comparer.levels, nonmatch_labels)
    return _train.make_weights(comparer, ms, us)
